import tempfile
from .base import *  # noqa

from libfaketime import reexec_if_needed
reexec_if_needed()

##########
# Django #
##########
MEDIA_ROOT = tempfile.mkdtemp()
SECRET_KEY = 'i^p%p^(gdh##2@ll*t)3^)1k87yo62)ig2=@sjoj*kas#lk46z'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': os.environ.get('TEST_PG_NAME', 'munch'),
        'USER': os.environ.get('TEST_PG_USER', 'munch'),
        'PASSWORD': os.environ.get('TEST_PG_PASS', 'munch'),
        'HOST': os.environ.get('TEST_PG_HOST', 'localhost'),
        'PORT': os.environ.get('TEST_PG_PORT', 15432),
    }
}
CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": os.environ.get('REDIS_PORT', 'redis://{}:{}/1'.format(
            os.environ.get('TEST_REDIS_HOST', 'localhost'),
            os.environ.get('TEST_REDIS_PORT', 16379))),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        }
    }
}

##########
# Celery #
##########
BROKER_BACKEND = 'memory'
CELERY_ALWAYS_EAGER = True
CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

##########
# Global #
##########
APPLICATION_URL = 'http://test.munch.example.com'
BYPASS_DNS_CHECKS = True
RETURNPATH_DOMAIN = 'test.munch.example.com'

#############
# Campaigns #
#############
CAMPAIGNS['BYPASS_RECIPIENTS_MX_CHECK'] = True

###########
# Domains #
###########
DOMAINS['DKIM_KEY_CONTENT'] = '"v=DKIM1; k=rsa; t=y; p=CHANGEME"'

################
# Upload store #
################
UPLOAD_STORE['URL'] = 'http://munch.example.com'
UPLOAD_STORE['IMAGE_MAX_WIDTH'] = 600
UPLOAD_STORE['BACKEND'] = (
    'munch.apps.upload_store.backends.LocalFileSystemStorage')

############
# Contacts #
############
CONTACTS = {
    'MAX_BULK_CONTACTS': 10000,
    'EXPIRATIONS': {
        'contact_queues:double-opt-in': timedelta(days=7),
        'contact_queues:bounce-check': timedelta(hours=1),
        'contact_queues:consumed_lifetime': timedelta(days=7),
        'contact_queues:failed_lifetime': timedelta(days=7),
        'contact_lists:double-opt-in': timedelta(days=7),
        'contact_lists:bounce-check': timedelta(hours=1),
        'contact_lists:consumed_lifetime': timedelta(days=7),
        'contact_lists:failed_lifetime': timedelta(days=7)}
}



import django_filters
from django import forms
from django.core.validators import EMPTY_VALUES


class DRFFilterSet(django_filters.FilterSet):
    # django-filters default is 'o'
    # DRF default is 'ordering'
    order_by_field = 'ordering'

    def _default_ordering_field(self):
        if self._meta.order_by:
            order_field = self.form.fields[self.order_by_field]
            data = self.form[self.order_by_field].data
            ordered_value = None
            try:
                ordered_value = order_field.clean(data)
            except forms.ValidationError:
                pass
            if ordered_value in EMPTY_VALUES and self.strict:
                ordered_value = self.form.fields[
                    self.order_by_field].choices[0][0]

            return ordered_value



import logging

from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.http import HttpResponseNotAllowed
from django.views.decorators.cache import cache_control

from munch.core.utils import get_mail_by_identifier

from .utils import WebKey
from .models import LinkMap
from .tasks import create_track_record

log = logging.getLogger(__name__)

# 1px transparent gif
TRACKER_PIXEL = (
    b"\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff"
    b"\xff\xff\x00\x00\x00\x21\xf9\x04\x01\x00\x00\x00\x00\x2c"
    b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b")


def do_track_and_redirect(mail_identifier, link_identifier):
    link_map = LinkMap.objects.filter(
        identifier=link_identifier).only('link').first()
    if link_map:
        create_track_record.apply_async(kwargs={
            'identifier': mail_identifier,
            'kind': 'click',
            'properties': {'link': link_identifier}})
        return HttpResponseRedirect(link_map.link)
    return HttpResponse(
        'No URL found with this ID', content_type='text/html', charset='UTF-8')


@cache_control(no_cache=True, max_age=0)
def tracking_open(request, identifier):
    if request.method == 'GET':
        mail = get_mail_by_identifier(identifier, must_raise=False)
        if mail:
            create_track_record.apply_async(kwargs={
                'identifier': identifier,
                'kind': 'read', 'properties': {'source': 'pixel'}})
        return HttpResponse(TRACKER_PIXEL, content_type='image/gif')
    return HttpResponseNotAllowed(permitted_methods=['GET'])


def tracking_redirect(request, identifier, link_identifier):
    if request.method == 'GET':
        return do_track_and_redirect(identifier, link_identifier)
    return HttpResponseNotAllowed(permitted_methods=['GET'])


def web_tracking_open(request, msg):
    """ Tracks opening from the hosted version

    Not a view, called by hosted.views
    """
    token = request.GET.get('web_key', None)
    if token:
        identifier = WebKey(token).get_identifier()
        if identifier:
            create_track_record.apply_async(kwargs={
                'identifier': identifier,
                'kind': 'read', 'properties': {'source': 'browser'}})
            return get_mail_by_identifier(identifier)


def web_tracking_redirect(request, web_key, link_identifier):
    if request.method == 'GET':
        identifier = WebKey(web_key).get_identifier()
        return do_track_and_redirect(identifier, link_identifier)
    return HttpResponseNotAllowed(permitted_methods=['GET'])



import logging
import email
from email.utils import parseaddr

import django_fsm
from django.utils.timezone import now as utc_now
from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import ValidationError

from munch.core.mail.backend import Backend
from munch.core.utils.tasks import task_autoretry
from munch.core.utils import get_mail_by_identifier
from munch.core.mail.utils import parse_email_date
from munch.core.mail.utils.dsn import DSNParser
from munch.core.mail.utils.dsn import InvalidDSNError
from munch.core.mail.models import return_path_parser
from munch.core.mail.models import unsubscribe_parser
from munch.core.mail.exceptions import RejectForbiddenTransition
from munch.apps.optouts.models import OptOut

from .validators import validate_existing_mail_domain

from .models import Mail
from .models import MailStatus
from .models import Message


log = logging.getLogger(__name__)


@task_autoretry(
    autoretry_on=(Exception, ),
    default_retry_delay=60 * 3,
    max_retries=(2 * 7 * 24 * 60 * 60) / 180,
    retry_message='Error while trying to build email. Retrying')
def send_mail(mail_pk):
    m = Mail.objects.get(pk=mail_pk)
    try:
        # Check if the recipient's domain has a MX record
        validate_existing_mail_domain(m.recipient)
    except ValidationError as exc:
        # No MX record, set status to IGNORED
        log.info('[{}] Ignored recipient {}, MX check failed'.format(
            m.identifier, m.recipient))
        try:
            MailStatus.objects.create(
                mail=m, status=MailStatus.IGNORED,
                creation_date=utc_now(), raw_msg=exc)
        except django_fsm.TransitionNotAllowed as e:
            raise RejectForbiddenTransition(m, e)
        return 'Ignored {}'.format(m.pk)
    else:
        log.info('[{}] Queueing email for recipient {}'.format(
            m.identifier, m.recipient))
        # explicitly gets the backend asking for systematic DSN.
        backend = Backend(
            mailstatus_class_path='munch.apps.campaigns.models.MailStatus',
            build_envelope_task_path=(
                'munch.apps.campaigns.utils.get_envelope'),
            record_status_task_path=(
                'munch.apps.campaigns.utils.record_status'))
        backend.send_message(
            m.identifier, m.recipient, m.get_headers(), attempts=0)
        return 'Sent {}'.format(m.pk)


@task_autoretry(
    default_retry_delay=60 * 30, max_retries=6 * 24 * 5,
    autoretry_on=(Exception, ), acks_late=True)
def handle_dsn(original_dsn, report_headers):
    """
    Handle a delivery status notification idempotent

    @param headers : a python dict with all the headers of the DSN email.
    """
    ENVELOPE_FIELD = 'To'
    original_dsn = email.message_from_string(original_dsn)

    try:
        envelope = parseaddr(original_dsn[ENVELOPE_FIELD])[1]
    except ValueError:
        log.info('Reject unknown mail: {}'.format(
            original_dsn[ENVELOPE_FIELD]))
        return

    try:
        dsn_report = DSNParser(report_headers)
    except InvalidDSNError as e:
        log.info('Reject invalid DSN: {}'.format(str(e)))
        return

    date = utc_now()
    next_status = dsn_report.get_next_mailstatus()

    try:
        if not return_path_parser.is_valid(envelope):
            log.info('Reject invalid address format: {}'.format(envelope))
            return
        identifier = return_path_parser.get_uuid(envelope)
        mail = Mail.objects.get(identifier=identifier)
    except (Mail.DoesNotExist, ValueError):
        log.info('Reject unknown mail: {}'.format(envelope))
        return
    log.info('Logging new status {} for {}'.format(next_status, mail))
    try:
        ms, created = MailStatus.objects.get_or_create(
            mail=mail, creation_date=date, status=next_status, defaults={
                'raw_msg': dsn_report.msg,
                'status_code': dsn_report.esmtp_status
            })
    except django_fsm.TransitionNotAllowed as e:
        log.info(
            'Reject forbidden dsn status transition '
            '({}): {}\n{}'.format(next_status, mail.identifier, str(e)))
        return
    # If there is nothing left to send, change the state of the message !
    remaining = mail.message.mails.legit_for(
        mail.message).pending().count()
    if remaining == 0:
        mail.message.status = Message.SENT
        mail.message.save()


@task_autoretry(
    default_retry_delay=60 * 30, max_retries=6 * 24 * 5,
    autoretry_on=(Exception, ))
def handle_fbl(main_headers, original_headers, feedback_headers):
    """
    Handle a feedback-loop return (see ARF format)
    Creates the relevant OptOut object, if needed.

    @param main_headers     : headers of the incomming msg \
                                (rfc822), as a dict
    @param original_headers : headers of the original message \
                                (rfc822), as a dict
    @param feedback_headers : headers of the feedback-return \
                                (ARF), as a dict
    """
    main_headers = email.message_from_string(main_headers)

    DATE_FIELD = (main_headers, 'Date')
    ENVELOPE_FIELD = (original_headers, 'Return-Path')

    def get(tupl):
        src, field = tupl
        return src[field]

    try:
        date_str = get(DATE_FIELD)
        envelope = parseaddr(get(ENVELOPE_FIELD))[1]
    except KeyError as e:
        log.info('Invalid feedloop-back, missing field: {}'.format(str(e)))
        return
    except ValueError:
        log.info('Unknown mail: {}'.format(get(ENVELOPE_FIELD)))
        return
    try:
        date = parse_email_date(date_str)
    except ValueError:
        log.info('Invalid date: {}'.format(date_str))
        return

    if not return_path_parser.is_valid(envelope):
        log.info('Reject invalid address format: {}'.format(envelope))
        return

    try:
        mail = get_mail_by_identifier(
            identifier=return_path_parser.get_uuid(envelope))
    except ObjectDoesNotExist:
        log.info('Reject unknown mail: {}'.format(envelope))
        return

    log.info('Unsubscribing {} via fbl'.format(mail.recipient))
    OptOut.objects.create_or_update(
        creation_date=date, identifier=mail.identifier,
        author=mail.get_author(), category=mail.get_category(),
        address=mail.recipient, origin=OptOut.BY_FBL)


@task_autoretry(
    default_retry_delay=60 * 30, max_retries=6 * 24 * 5,
    autoretry_on=(Exception, ))
def handle_mail_optout(headers):
    """
    For the cases a user (or a machine) sends a message to the
    List-Unsubscribe email.

    We do not rely on the sender address at all
    (think about aliases & so on...)

    @param headers the headers of the unsubscription email.
    """
    DATE_FIELD = 'Date'
    UNSUBSCRIBE_ADDR = 'To'

    try:
        date_str = headers[DATE_FIELD]
        unsubscribe_addr = parseaddr(headers[UNSUBSCRIBE_ADDR])[1]
    except KeyError as e:
        log.info('Invalid optout, missing field: {}'.format(str(e)))
        return
    except ValueError:
        log.info('Unknown mail: {}'.format(headers[UNSUBSCRIBE_ADDR]))
        return

    try:
        date = parse_email_date(date_str)
    except ValueError:
        log.info('Invalid date: {}'.format(date_str))
        return

    if not unsubscribe_parser.is_valid(unsubscribe_addr):
        log.info(
            'Reject invalid unsubscribe address format: {}'.format(
                unsubscribe_addr))
        return

    envelope_from = return_path_parser.new(
        unsubscribe_parser.get_uuid(unsubscribe_addr))

    try:
        mail = Mail.objects.get(
            identifier=unsubscribe_parser.get_uuid(unsubscribe_addr))
    except ObjectDoesNotExist:
        log.info('Reject unknown mail: {}'.format(envelope_from))
        return

    existing_optout = OptOut.objects.filter(
        identifier=mail.identifier, origin=OptOut.BY_MAIL).only(
        'pk', 'creation_date')
    if existing_optout:
        optout = existing_optout[0]
        log.info('Already unsubscribed: {} via mail at {} (pk={})'.format(
            mail.recipient, optout.creation_date, optout.pk))
    else:
        log.info('Unsubscribing {} via mail'.format(mail.recipient))
        OptOut.objects.create_or_update(
            author=mail.get_author(), category=mail.get_category(),
            creation_date=date, identifier=mail.identifier,
            address=mail.recipient, origin=OptOut.BY_MAIL)



import logging

from munch.core.utils.tasks import task_autoretry
from munch.core.mail.exceptions import SoftFailure
from munch.core.mail.exceptions import HardFailure
from munch.core.mail.exceptions import RejectUnknownIdentifier

from .models import Mail

log = logging.getLogger(__name__)


def get_envelope(identifier, **kwargs):
    try:
        return Mail.objects.get(identifier=identifier).as_envelope()
    except Mail.DoesNotExist as exc:
        raise SoftFailure(exc)


def get_message_from_identifier(identifier, *kwargs):
    try:
        return Mail.objects.get(identifier=identifier).as_message()
    except Mail.DoesNotExist as exc:
        raise SoftFailure(exc)


def get_mail_or_raise(identifier):
    """ Fetch the http stored returnpath

    See http_returnpath policy for the inverse action (storing)

    :param: X-CM-Message-Id
    :return: MailMetadata
    """

    try:
        mail = Mail.objects.get(identifier=identifier)
    except Mail.DoesNotExist:
        raise RejectUnknownIdentifier(identifier=identifier)
    return mail


@task_autoretry(
    max_retries=None, autoretry_on=(Exception, ),
    autoretry_exclude=(SoftFailure, HardFailure),
    retry_message='Error while trying to record_status. Retrying.')
def record_status(mailstatus, identifier, relay_ehlo, reply=None):
    """ Save MailStatus and send SMTP Status task """
    log.debug('[{}] Recording "{}" status...'.format(
        identifier, mailstatus.status))
    try:
        mail_metadata = get_mail_or_raise(identifier)
        mailstatus.mail = mail_metadata
        mailstatus.save()
    except (SoftFailure, RejectUnknownIdentifier) as exc:
        raise SoftFailure(exc)



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "munch-core"
__summary__ = "Munch is an emailing platform."
__uri__ = "https://github.com/crunchmail/munch-core"

__version__ = "0.1.0"

__author__ = u"Crunchmail"
__email__ = "dev@crunchmail.com"

__license__ = "GNU AGPLv3"
__copyright__ = "Copyright 2016 %s" % __author__



from django.conf.urls import url
from django.conf.urls import include

from munch.core.utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register(
    'organizations', views.OrganizationViewSet, base_name='organization')
router.register('users', views.MunchUserViewSet, base_name='munchuser')
router.register(
    'applications/api', views.APIApplicationViewSet,
    base_name='applications-api')
router.register(
    'applications/smtp', views.SmtpApplicationViewSet,
    base_name='applications-smtp')

urlpatterns = [
    url(
        '^users/(?P<user_pk>\d+)/regen_secret/$',
        views.RegenSecretView.as_view()),
    url(
        '^users/(?P<user_pk>\d+)/change_password/$',
        views.ChangePasswordView.as_view()),

    # Following urls are outside of r'^users/'
    url('^me/$', views.MeAPIView.as_view(), name='me'),
    url(r'^auth/password_reset/$', views.PasswordResetInitView.as_view(),
        name='api_password_reset_init'),
    url((
        r'^auth/password_reset/change/(?P<uidb64>[0-9A-Za-z_\-]+)'
        r'/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'),
        views.PasswordResetSetView.as_view(),
        name='api_password_reset_set'),

] + router.urls

# We do not use prefix because it's impossible for DRF to register a ViewSet
# at index of Router. And some urls from users app are outside of r'^users/'
api_urlpatterns_v1 += [url('', include(urlpatterns, namespace='users'))]



import sys

from celery.signals import celeryd_after_setup

from munch.core.utils import get_worker_type
from munch.core.celery import catch_exception
from munch.core.celery import munch_tasks_router


def register_tasks():
    tasks_map = {
        'core': ['munch.apps.campaigns.tasks.send_mail'],
        'status': [
            'munch.apps.campaigns.tasks.handle_dsn',
            'munch.apps.campaigns.tasks.handle_fbl',
            'munch.apps.campaigns.tasks.record_status',
            'munch.apps.campaigns.tasks.handle_mail_optout',
        ]
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'campaigns')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    if get_worker_type() in ['status', 'all']:
        from .tasks import handle_dsn  # noqa
        from .tasks import handle_fbl  # noqa
        from .utils import record_status  # noqa
        from .tasks import handle_mail_optout  # noqa
        sys.stdout.write('[campaigns-app] Registering worker as STATUS...')
        munch_tasks_router.register_as_worker('status')

    if get_worker_type() in ['core', 'all']:
        from .tasks import send_mail  # noqa
        sys.stdout.write('[campaigns-app] Registering worker as CORE...')
        munch_tasks_router.register_as_worker('core')



from .base import *
from datetime import timedelta
from urllib.parse import urlparse

##########
# Django #
##########
INSTALLED_APPS += [
    # 'raven.contrib.django.raven_compat',
    # 'munch_storage',
    # Other
    # 'debug_toolbar',
    'munch_dentifrice'
]
# MIDDLEWARE_CLASSES.insert(MIDDLEWARE_CLASSES.index('django.middleware.common.CommonMiddleware') - 1, 'corsheaders.middleware.CorsMiddleware')  # noqa
# CORS_ORIGIN_ALLOW_ALL = True
DEBUG = True
SECRET_KEY = 'test'
CONTAINER_IP = urlparse(os.environ.get(
    'DOCKER_HOST', 'tcp://localhost')).hostname
BROKER_URL = 'amqp://guest:guest@{}:5682/munch'.format(CONTAINER_IP)
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'munch',
        'USER': 'munch',
        'PASSWORD': 'munch',
        'HOST': CONTAINER_IP,
        'PORT': '15432',
    }
}

CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": "redis://{}:16379/1".format(CONTAINER_IP),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        }
    }
}

# MIDDLEWARE_CLASSES += ['munch.settings.local.NonHtmlDebugToolbarMiddleware']


class NonHtmlDebugToolbarMiddleware:
    """
    The Django Debug Toolbar usually only works for views that return HTML.
    This middleware wraps any non-HTML response in HTML if the request
    has a 'debug' query parameter (e.g. http://localhost/foo?debug)
    Special handling for json (pretty printing) and
    binary data (only show data length)
    """

    @staticmethod
    def process_response(request, response):
        import json
        from django.http import HttpResponse

        if request.GET.get('debug') == '':
            if response['Content-Type'] == 'application/octet-stream':
                new_content = '<html><body>Binary Data, ' \
                    'Length: {}</body></html>'.format(len(response.content))
                response = HttpResponse(new_content)
            elif response['Content-Type'] != 'text/html':
                content = response.content
                try:
                    json_ = json.loads(content.decode('utf-8'))
                    content = json.dumps(json_, sort_keys=True, indent=2)
                except ValueError:
                    pass
                response = HttpResponse('<html><body><pre>{}'
                                        '</pre></body></html>'.format(content))

        return response

###########
# Logging #
###########
"""
RAVEN_CONFIG = {
    'dsn': 'http://13a320fda9ef4ae39c7a9b7f724859ef:6547c908892548a3863c379398f24090@localhost:9000/2',
}
"""
LOGGING['loggers']['munch'] = {
    'handlers': ['console'],
    'filters': [],
    'propagate': False,
    'level': logging.DEBUG,
}
LOGGING['loggers']['munch_mailsend'] = {
    'handlers': ['console'],
    'level': logging.DEBUG,
    'propagate': False
}
"""
LOGGING['disable_existing_loggers'] = True
LOGGING['handlers']['sentry'] = {
    'level': 'ERROR',
    'class': 'raven.contrib.django.raven_compat.handlers.SentryHandler',
}
LOGGING['loggers']['root'] = {
    'handlers': ['console', 'sentry'],
    'level': 'WARNING',
}
LOGGING['loggers']['raven'] = {
    'level': 'DEBUG',
    'handlers': ['console'],
    'propagate': False,
}
LOGGING['loggers']['sentry.errors'] = {
    'level': 'DEBUG',
    'handlers': ['console'],
    'propagate': False,
}
"""

##########
# Global #
##########
APPLICATION_URL = 'http://localhost:8000'
MASS_SENDER_DOMAIN = 'localhost'
BYPASS_DNS_CHECKS = True
MASS_EMAIL_BACKEND = 'munch_mailsend.backend.Backend'

##################
# Custom headers #
##################
X_POOL_HEADER = 'X-CM-Pool'

##########
# Celery #
##########
CELERYBEAT_SCHEDULE.update({
    'ping_workers': {
        'task': 'mailsend.tasks.ping_workers',
        'schedule': timedelta(seconds=30),
    },
    'check_disabled_workers': {
        'task': 'mailsend.tasks.check_disabled_workers',
        'schedule': timedelta(seconds=30),
    }
})

#################
# Transactional #
#################
TRANSACTIONAL['PROXYPROTO_ENABLED'] = False
TRANSACTIONAL['SMTP_BIND_HOST'] = '0.0.0.0'
TRANSACTIONAL['SMTP_BIND_PORT'] = 1035
TRANSACTIONAL['SMTP_SMARTHOST_TLS'] = {
    'keyfile': os.path.join(
        BASE_DIR, '../../utils/ssl/postfix.example.com.key.pem'),
    'certfile': os.path.join(
        BASE_DIR, '../../utils/ssl/postfix.example.com.cert.pem'),
}
TRANSACTIONAL['HEADERS_TO_REMOVE'] += [X_POOL_HEADER]


############
# Mailsend #
############
MAILSEND = {
    # All Internal mailsend emails will be send to a blackhole
    'SANDBOX': False,
    'SMTP_WORKER_FORCE_MX': [{
        'domain': 'example.com',
        'destination': CONTAINER_IP, 'port': '2525'
    }],
    'RELAY_TIMEOUTS': {
        'connect_timeout': 30.0, 'command_timeout': 30.0,
        'data_timeout': None, 'idle_timeout': None},
    # Timeout for MailStatus cache
    'MAILSTATUS_CACHE_TIMEOUT': 60 * 60 * 24 * 15,
    'X_USER_ID_HEADER': X_USER_ID_HEADER,
    'X_POOL_HEADER': X_POOL_HEADER,
    'X_MESSAGE_ID_HEADER': X_MESSAGE_ID_HEADER,
    # Letting to None will make it use the host FQDN
    'SMTP_WORKER_EHLO_AS': 'localhost',
    # Letting to None fallback to system routing
    # example: '1.2.3.4'
    'SMTP_WORKER_SRC_ADDR': '127.0.0.1',
    # Backoff time is exponentially growing up to max_retry_interval and then
    # staying there on each retry till we reach time_before_drop.
    'RETRY_POLICY': {
        # Minimun time between two retries
        'min_retry_interval': 600,
        # Maximum time between two retries
        'max_retry_interval': 3600,
        # Time before we drop the mail and notify sender
        'time_before_drop': 2 * 24 * 3600},
    'BINARY_ENCODER': None,
    'BLACKLISTED_HEADERS': [
        X_POOL_HEADER,
        X_HTTP_DSN_RETURN_PATH_HEADER,
        X_SMTP_DSN_RETURN_PATH_HEADER],
    'RELAY_POLICIES': [
        'munch.apps.transactional.policies.relay.headers.RewriteReturnPath',
        'mailsend.policies.relay.headers.StripBlacklisted',
        'mailsend.policies.relay.dkim.Sign'],
    'WORKER_POLICIES': [
        'mailsend.policies.mx.pool.Policy',
        'mailsend.policies.mx.rate_limit.Policy',
        'mailsend.policies.mx.greylist.Policy',
        'mailsend.policies.mx.warm_up.Policy'],
    'WORKER_POLICIES_SETTINGS': {
        'rate_limit': {
            'domains': [
                (r'.*', 2)],
            'max_queued': 60 * 15},
        'warm_up': {
            'prioritize': 'equal',
            'domain_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'days_watched': 10,
            },
            'ip_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'enabled': False,
                'days_watched': 10,
            }
        },
        'pool': {'pools': ['default']},
        'greylist': {'min_retry': 60 * 5},
    },
    'WARM_UP_DOMAINS': {},
    'DKIM_PRIVATE_KEY': """-----BEGIN RSA PRIVATE KEY-----
MIICXQIBAAKBgQDiTemlcIMSRZ3us+cSt8Z9SjJ3H49FPq2BoT6IIbYX0PGj3jhu
HsyqX+NLpjCgtKAWMw4ik3sjTkTz56NRxOavvzif4gTSg4/d6VmHs5Mw4/HSjMrZ
9+vDhC3Wt5heQSOvyIwKj8+KgyGRRRmoZbmDN4ms0Gq1OFIHm4T7mQP28wIDAQAB
AoGAEbpkyU8NFYtame6B9Zdr9ziux2IziQsl8He/PE7XwvndVCb+aLIE+nvUhIKa
YJyFxfdt7gt7pAJnqGvHAYrZP1mKsZDIkEgFruYwNVGCLkIQDk3RJwNb3Qpgwskj
mJnQVLZOE120NxBXqrIvNz0Gi61H08gUB5TFKUM3bFxc38ECQQD9X+kmwElyc+Oa
EYJIO6NuvVoDJQBZX+tuaUuomAbqejoMgo5njdFKLR+IqBO43lG9FAitgWCanyvh
S/eoGqKRAkEA5KYyUdl5lHOxHcLo6VKT7Y9nZBPoWn0QhMYuN6vJDpzEr9S+4sSH
CCbawaJGxrJFIiA4vBodOjvIQwOur1Q7QwJBAOHzwiv4locmqfYfXxujc5+x5K+h
M6qAS6fu5rW2vZQk49d8Jhpa8iVAEDsCCHR4blQ7pXF1Sv0YrT0BTh3vgsECQB/q
gwOh07LBI2wAFPrcqAF1Dv2NOdXHt1KRR0pGFF6Ry3Kvw6VrwV2F7uswd6isobHN
xZ2cF5BVX/LaxLt8inkCQQCFKnTfKaEp4W90zqBUWZn+M6zRZfyi+2gN95Srikra
DxEoQHGETi2ZDSihkvx/spwT3r8kNvfTYXpgwt14l0jq
-----END RSA PRIVATE KEY-----""",
    'DKIM_SELECTOR': "test",
    'TLS': {
        'keyfile': os.path.join(
            BASE_DIR, '../../utils/ssl/postfix.example.com.key.pem'),
        'certfile': os.path.join(
            BASE_DIR, '../../utils/ssl/postfix.example.com.cert.pem')},
}

#############
# Campaigns #
#############
CAMPAIGNS['SKIP_SPAM_CHECK'] = True
CAMPAIGNS['SKIP_VIRUS_CHECK'] = True
CAMPAIGNS['BYPASS_RECIPIENTS_MX_CHECK'] = True

############
# Contacts #
############
CONTACTS = {
    # How many contacts can we add in a single API request ?
    'MAX_BULK_CONTACTS': 10000,
    'EXPIRATIONS': {
        'contact_queues:double-opt-in': timedelta(days=7),
        'contact_queues:bounce-check': timedelta(hours=1),
        'contact_queues:consumed_lifetime': timedelta(days=7),
        'contact_queues:failed_lifetime': timedelta(days=7),
        'contact_lists:double-opt-in': timedelta(days=7),
        'contact_lists:bounce-check': timedelta(hours=1),
        'contact_lists:consumed_lifetime': timedelta(days=7),
        'contact_lists:failed_lifetime': timedelta(days=7)}
}

################
# Upload store #
################
UPLOAD_STORE['URL'] = 'http://munch.example.com'
UPLOAD_STORE['IMAGE_MAX_WIDTH'] = 600
UPLOAD_STORE['BACKEND'] = (
    'munch.apps.upload_store.backends.LocalFileSystemStorage')

#########
# Spamd #
#########
SPAMD_PORT = 1783

#########
# Clamd #
#########
CLAMD_PORT = 13310

###############
# Backmuncher #
###############
BACKMUNCHER['PROXYPROTO_ENABLED'] = True



from django.test import TestCase
from django.db import IntegrityError

from .factories import UserFactory
from .factories import APIApplicationFactory
from .factories import SmtpApplicationFactory


class APIApplicationTestCase(TestCase):
    def test_create_application(self):
        application = APIApplicationFactory()

        secret = application.secret

        self.assertTrue(application.secret)
        self.assertTrue(application.author)
        self.assertTrue(application.identifier)

        application.regen_secret()

        self.assertNotEqual(secret, application.secret)

        with self.assertRaises(IntegrityError):
            APIApplicationFactory(
                author=application.author, identifier=application.identifier)


class SmtpApplicationTestCase(TestCase):
    def test_create_application(self):
        application = SmtpApplicationFactory()

        username = application.username
        secret = application.secret

        self.assertTrue(application.secret)
        self.assertTrue(application.username)
        self.assertTrue(application.author)
        self.assertTrue(application.identifier)

        application.regen_credentials()
        application.save()

        self.assertNotEqual(secret, application.secret)
        self.assertNotEqual(username, application.username)

        with self.assertRaises(IntegrityError):
            SmtpApplicationFactory(username=application.username)

        with self.assertRaises(IntegrityError):
            SmtpApplicationFactory(
                author=application.author, identifier=application.identifier)



from rest_framework import serializers

from ...models import Image


class ImageSerializer(serializers.HyperlinkedModelSerializer):
    expiration = serializers.DurationField(required=False)
    url = serializers.HyperlinkedIdentityField(
        view_name='v1:upload-store:image-detail',
        lookup_field='hash', lookup_url_kwarg='pk')

    class Meta:
        model = Image
        fields = ('url', 'file', 'width', 'upload_date', 'expiration')
        read_only_fields = ('upload_date',)

    def create(self, valid_data):
        organization = self.context['request'].user.organization
        valid_data['organization'] = organization
        return super().create(valid_data)



import logging

from django.apps import apps
from django.contrib.auth.models import Group
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand

log = logging.getLogger(__name__)

MIGRATION = """# -*- coding: utf-8 -*-
# Generated by "permissions_migration" management command
from django.db import migrations
from django.core.management.sql import emit_post_migrate_signal

{raw}

def update_content_types(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    emit_post_migrate_signal(False, 'default', db_alias)


def load_permissions(apps, schema_editor):
    Group = apps.get_model('auth', 'group')
    Permission = apps.get_model('auth', 'permission')
    ContentType = apps.get_model('contenttypes', 'contenttype')

    # Delete previous permissions
    for model in PERMISSIONS:
        content_type = ContentType.objects.get(
            app_label='{app_label}', model=model)
        Permission.objects.filter(content_type=content_type).delete()

    # Load permissions
    for model_name, permissions in PERMISSIONS.items():
        for permission_codename, permission_name in permissions:
            content_type = ContentType.objects.get(
                app_label='{app_label}', model=model_name)
            if not Permission.objects.filter(
                    codename=permission_codename,
                    content_type=content_type).exists():
                Permission.objects.create(
                    name=permission_name,
                    codename=permission_codename,
                    content_type=content_type)

    # Group permissions
    for group_name, models in GROUP_PERMISSIONS.items():
        group, _ = Group.objects.get_or_create(name=group_name)
        for model_name, permissions in models.items():
            content_type = ContentType.objects.get(
                app_label='{app_label}', model=model_name)
            for codename in permissions:
                permission, _ = Permission.objects.get_or_create(
                    codename=codename, content_type_id=content_type.id)
                group.permissions.add(permission)
                group.save()


class Migration(migrations.Migration):
    # TODO: Add your dependencies here then remove "raise"add:
    #   dependencies = [('{app_label}', '0001_initial')]
    raise Exception('You forgot to edit "dependencies".')
    dependencies = [('{app_label}', '0001_initial')]
    operations = [
        migrations.RunPython(update_content_types, reverse_code=None),
        migrations.RunPython(load_permissions, reverse_code=None)]
"""


class Command(BaseCommand):
    help = "Get a permissions migration template."

    def add_arguments(self, parser):
        parser.add_argument('app_label', type=str, help='App label.')

    def handle(self, *args, **options):
        app_label = options.get('app_label')
        raw = "PERMISSIONS = {\n"
        for model in apps.get_app_config(app_label).get_models():
            model_name = model._meta.model_name
            raw += "    '{}': [".format(model_name)
            content_type = ContentType.objects.get(
                app_label=app_label, model=model_name)
            for permission in Permission.objects.filter(
                    content_type=content_type):
                if not permission.name:
                    log.info(
                        '"{}.{}.{}" (pk:{}) permission does '
                        'not have name.'.format(
                            app_label, model_name,
                            permission.codename, permission.pk))
                if not permission.codename:
                    log.warning(
                        '"{}.{}.{}" (pk:{}) permission does '
                        'not have codename.'.format(
                            app_label, model_name,
                            permission.codename, permission.pk))
                raw += "\n        ('{}', '{}'), ".format(
                    permission.codename.replace('\'', '\\\''),
                    permission.name.replace('\'', '\\\''))
            raw += "],\n"
        raw += "}\n"
        raw += "GROUP_PERMISSIONS = {\n"
        for group in Group.objects.all():
            raw += "    '{}': {{\n".format(group.name)
            for model in apps.get_app_config(app_label).get_models():
                model_name = model._meta.model_name
                raw += "        '{}': [".format(model_name)
                content_type = ContentType.objects.get(
                    app_label=app_label, model=model_name)
                for permission in group.permissions.filter(
                        content_type=content_type):
                    raw += "\n            '{}', ".format(permission.codename)
                raw += "],\n"
            raw += "    },\n"
        raw += "}\n"
        self.stdout.write(MIGRATION.format(app_label=app_label, raw=raw))



from django.conf import settings
from django.db import connection
from django.db import transaction
from rest_framework import fields
from rest_framework import serializers
from rest_framework.settings import api_settings
from rest_framework.serializers import ValidationError
from rest_framework.serializers import DjangoValidationError
from rest_framework.serializers import get_validation_error_detail

from munch.core.utils.serializers import HALLinksField
from munch.core.utils.serializers import SizeLimitedListSerializer
from munch.core.utils.relations import ScopedHyperLinkedRelatedField

from ...models import Contact
from ...models import ContactList
from ...models import ContactQueue
from ...models import CollectedContact
from ...models import ContactListPolicy
from ...models import ContactQueuePolicyAttribution
from ...validators import properties_schema_validator
from .permissions import ContactListMergePermission


class QueuePolicySerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = ContactListPolicy
        fields = ['url', 'name', ]
        read_only_fields = ['url']
        extra_kwargs = {
            'url': {'view_name': 'contacts:contactlistpolicy-detail'}}


class ContactQueueSerializer(serializers.HyperlinkedModelSerializer):
    policies = serializers.HyperlinkedRelatedField(
        many=True, view_name='contacts:contactlistpolicy-detail',
        queryset=ContactListPolicy.objects)
    _links = HALLinksField(
        nested_endpoints=['consume'],
        view_name='contacts:contactqueue-detail')
    subscription = serializers.ReadOnlyField()
    contact_fields = serializers.ListField(
        child=serializers.DictField(), required=False,
        validators=[properties_schema_validator])

    class Meta:
        model = ContactQueue
        fields = [
            'url', 'policies', 'properties', 'contact_fields',
            'contacts_count', '_links', 'subscription']
        read_only_fields = ['url', 'contacts_count', '_links', 'subscription']
        extra_kwargs = {'url': {'view_name': 'contacts:contactqueue-detail'}}

    def create(self, validated_data):
        with transaction.atomic():
            data = {k: v for k, v in validated_data.items() if k not in
                    ('policies',)}
            queue = ContactQueue(**data)
            queue.save()
            for policy in validated_data['policies']:
                pa = ContactQueuePolicyAttribution(
                    contact_queue=queue, policy=policy)
                pa.save()
            return queue

    def update(self, instance, validated_data):
        with transaction.atomic():
            for k, v in validated_data.items():
                if k != 'policies':
                    setattr(instance, k, v)
            instance.save()

            # handle the through model, not that easy, see
            # https://docs.djangoproject.com/en/dev/topics/db/models/#intermediary-manytomany
            instance.policies.clear()
            for policy in validated_data['policies']:
                ContactQueuePolicyAttribution.objects.create(
                    contact_queue=instance, policy=policy)
            return instance

    def validate_contact_fields(self, value):
        ContactFieldsFieldHelper.validate(
            self.instance, value, reverse='collected_contacts')
        return value


class CollectedContactSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = CollectedContact
        fields = ['address', 'properties', 'creation_date', 'update_date']

    def to_representation(self, obj):
        data = super().to_representation(obj)
        data['properties'] = PropertiesFieldHelper.to_representation(
            obj, reverse="contact_queue")
        return data

    def to_internal_value(self, data):
        obj = super().to_internal_value(data)
        try:
            obj['properties'] = PropertiesFieldHelper.to_internal_value(
                obj, reverse="contact_queue")
        except ValidationError as err:
            raise ValidationError({'properties': err.detail})

        return obj


class ContactQueueDetailSerializer(serializers.HyperlinkedModelSerializer):
    policies = QueuePolicySerializer(many=True, read_only=True)
    contacts = CollectedContactSerializer(many=True, read_only=True)

    class Meta:
        model = ContactQueue
        fields = [
            'url', 'policies', 'properties', 'contacts_count', 'contacts']
        read_only_fields = [
            'url', 'policies', 'properties', 'contacts_count', 'contacts']
        extra_kwargs = {'url': {'view_name': 'contacts:contactqueue-detail'}}


class URLRelatedDefault:
    """ Fills a field from the request URL
    """
    def __init__(self, url_pk, target_model, *args, **kwargs):
        self.url_pk = url_pk
        self.target_model = target_model

    def set_context(self, serializer_field):
        kwargs = serializer_field.context['kwargs']
        self.related_obj = self.target_model.objects.get(
            pk=kwargs[self.url_pk])

    def __call__(self):
        return self.related_obj


class ContactListSerializer(serializers.HyperlinkedModelSerializer):
    _links = HALLinksField(
        nested_endpoints=['contacts', 'merge'],
        view_name='contacts:contactlist-detail')

    # need to include it here so that unique_together in checked and handled
    # properly with a 400 error
    author = serializers.HiddenField(default=serializers.CreateOnlyDefault(
        serializers.CurrentUserDefault()))
    contact_fields = serializers.ListField(
        child=serializers.DictField(), required=False,
        validators=[properties_schema_validator])
    subscription = serializers.ReadOnlyField()

    class Meta:
        model = ContactList
        fields = [
            'url', 'name', 'contact_fields', 'properties',
            'source_type', 'source_ref', 'subscription',
            'contacts_count', '_links', 'author']
        read_only_fields = ['url', 'contacts_count', 'author']
        extra_kwargs = {'url': {'view_name': 'contacts:contactlist-detail'}}

    def validate_contact_fields(self, value):
        ContactFieldsFieldHelper.validate(self.instance, value)
        return value


class ContactBulkSerializer(SizeLimitedListSerializer):
    class Meta:
        # make it lazy, otherwise, overrided settings value can be missed
        max_items = lambda: settings.CONTACTS['MAX_BULK_CONTACTS']

    def create(self, validated_data):
        contacts = [Contact(**item) for item in validated_data]
        return Contact.objects.bulk_create(contacts, batch_size=30000)

    def to_internal_value_error(self, data):
        """ Returns both errors and valid data
        :returns: a couple: list of valid values
                            ValidationError with all errors
        """
        # Called 1 time for all the data list

        if fields.html.is_html_input(data):
            data = fields.html.parse_html_list(data)

        if not isinstance(data, list):
            message = self.error_messages['not_a_list'].format(
                input_type=type(data).__name__
            )
            raise ValidationError({
                api_settings.NON_FIELD_ERRORS_KEY: [message]
            })

        ret = []
        errors = {
            'validation_errors': {},
            'no_address': 0,
            'duplicates': []
        }
        unique_couples = set()

        for item in data:
            try:
                validated = self.child.run_validation(item)
            except ValidationError as exc:
                if not item.get('address'):
                    errors['no_address'] += 1
                else:
                    errors['validation_errors'][
                        item.get('address')] = exc.detail
            else:
                unicity = (validated['contact_list'], validated['address'])
                if unicity not in unique_couples:
                    ret.append(validated)
                    unique_couples.add(unicity)
                else:
                    errors['duplicates'].append(item.get('address'))

        if any(v for v in errors.values()):
            return ret, ValidationError(errors)

        return ret, ValidationError({})

    def is_partly_invalid(self):
        """ Allows to have a multiple serializer (many=True) with part of the
            raise serializers data valid and part of the data invalid.
        """
        assert hasattr(self, 'initial_data'), (
            'Cannot call `.is_partly_valid()`no `data=` keyword argument was'
            'passed when instantiating the serializer instance.'
        )
        has_errors = False

        # this one may trigger a ValidationError up to http client
        self.run_list_size_validation(self.initial_data)

        # those can't, from here, validation is always considered good
        if not hasattr(self, '_validated_data'):
            self._validated_data, exc = self.run_partial_validation(
                self.initial_data)
            self._partial_errors = exc.detail
            if exc.detail:
                has_errors = True

        # partly valid if at least something is valid
        return has_errors

    def run_partial_validation(self, data=fields.empty):
        """
        We override the default `run_validation`, because the validation
        performed by validators and the `.validate()` method should
        be coerced into an error dictionary with a 'non_fields_error' key.
        """
        # Called 1 time for all the data list
        # output of this function is set into self._validated_data by is_valid
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data
        value, error = self.to_internal_value_error(data)
        try:
            self.run_validators(value)  # does nothing in our case
            value = self.validate(value)
            assert value is not None, \
                '.validate() should return the validated data'
        except (ValidationError, DjangoValidationError) as exc:
            raise ValidationError(detail=get_validation_error_detail(exc))
        return value, error


class ContactFieldsFieldHelper:
    @classmethod
    def validate(cls, instance, new_contact_fields, reverse='contacts'):
        errors = {}
        # Create reversed contact_fields like that:
        # {
        #  "Age": {"type": "Integer", "required": True},
        #  "Firstname": {"type": "Char", "required": True}
        # }

        # Raise error if new field have same name than another
        _duplicates = {}
        for field in new_contact_fields:
            field = field['name'].strip().lower()
            if field not in _duplicates:
                _duplicates[field] = 0
            _duplicates[field] += 1
        for field, count in _duplicates.items():
            if count > 1:
                if not errors.get(field):
                    errors[field] = []
                errors[field].append(
                    "can't add same field multiple times (case insensitive)")
        del _duplicates

        reversed_initial = {}
        # Check if instance a new one or an update
        if hasattr(instance, 'pk'):
            reversed_initial = {
                field.get('name'): {
                    'type': field.get('type'),
                    'required': field.get('required', False)}
                for field in instance.contact_fields}
        reversed_new = {
            field.get('name').strip().lower(): {
                'type': field.get('type'),
                'required': field.get('required', False)}
            for field in new_contact_fields}
        for field_name, field in reversed_new.items():
            if not errors.get(field_name):
                errors[field_name] = []
            field_type = field.get('type')

            # Raise error if new field have same name than another
            lowered_fields = [f.lower() for f in list(reversed_new)]
            if field_name.lower() in [
                    x for x in list(reversed_new)
                    if lowered_fields.count(x.lower()) > 1]:
                errors[field_name].append(
                    "can't add same field multiple times (case insensitive)")
                continue

            # Do not check field "updates" if it's a new Contact(Queue/List)
            if not instance:
                continue

            # Raise error if new field is required and there are contacts
            if field_name not in reversed_initial and field.get('required'):
                if getattr(instance, reverse).all().count():
                    errors[field_name].append("can't add field as required")
                    continue
            # Check if initial field has been updated
            if field_name in reversed_initial:
                initial_field = reversed_initial.get(field_name)
                # Raise error if existing field has changed its type
                if initial_field.get('type') != field_type:
                    # If new type is not Char we must raise an error
                    if field_type != "Char":
                        errors[field_name].append(
                            "can't change existing field type to %s"
                            % field_type)
                # Raise error if existing field setted as required
                if not initial_field.get('required') and field.get('required'):
                    _filter = {'properties__%s__isnull' % field_name: True}
                    # Check if it's possible to set existing field as required
                    # by checking if every contacts field is not empty
                    if getattr(instance, reverse).filter(**_filter).exists():
                        errors[field_name].append(
                            "can't set existing field as required")
                        continue
            # Iterate over removed fields
            removed_fields = reversed_initial.keys() - reversed_new.keys()
            # Clean all contacts properties if there are removed fields
            if removed_fields:
                db_table = getattr(instance, reverse).model._meta.db_table
                contact_ids = getattr(instance, reverse).values_list(
                    "pk", flat=True)
                # Run cleaning only if there are contacts attached
                if contact_ids:
                    cursor = connection.cursor()
                    # Need to format just db_table alone because parametrized
                    # statements seems to be only for "values" and not table
                    # names. See: http://stackoverflow.com/a/9354420
                    # This raw SQL statement use Postgres `delete` method to
                    # manipulate hstore that we can do that with Django.
                    cursor.execute(("""
                        UPDATE {db_table}
                        SET properties = delete(properties, %s)
                        WHERE id in %s""".format(db_table=db_table)), [
                        list(removed_fields), tuple(contact_ids)])

        if any([error for error in errors.values()]):
            raise ValidationError({k: v for k, v in errors.items() if v})

        return new_contact_fields


class PropertiesFieldHelper:
    @classmethod
    def get_field_serializer(cls, field_type):
        try:
            # Try to retrieve right field type
            serializer = getattr(serializers, '%sField' % field_type)
        except AttributeError:
            # Fallback on CharField
            serializer = serializers.CharField
        return serializer

    @classmethod
    def to_representation(cls, instance, reverse="contact_list"):
        serialized_properties = {}
        properties = instance.properties
        # Iterate over Contact.list.contact_fields
        for field in getattr(instance, reverse).contact_fields:
            field_name, field_type = field.get('name'), field.get('type')
            field_value = properties.get(field_name)
            # If field is empty and not required let's ignore it
            if not field_value and not field.get('required'):
                continue
            field_serializer = cls.get_field_serializer(field_type)()
            try:
                serialized_properties[field_name] = field_serializer\
                    .to_representation(
                        field_serializer.to_internal_value(field_value))
            except (TypeError, ValueError, ValidationError) as err:
                # If we go here it's because we validate a field in
                # `to_internal_value` that we are unable to representate.
                # Which is not a good point.
                errors = ['Error while rendering field value']
                if isinstance(err, ValidationError):
                    errors += err.detail
                else:
                    errors.append(str(err))
                raise Exception({field_name: errors})
        return serialized_properties

    @classmethod
    def to_internal_value(cls, data, reverse="contact_list"):
        """
            reverse parameter is attribut to
            retrieve: ContactList or ContactQueue
        """
        errors = {}
        returned_properties = {}
        properties = data.get('properties')
        # Iterate over Contact.list.contact_fields
        for field in data.get(reverse).contact_fields:
            field_name, field_type = field.get('name'), field.get('type')
            field_value = properties.get(field_name)
            errors[field_name] = []
            # If field is empty and required, raise an error and continue
            if not field_value and field.get('required'):
                errors[field_name].append('is required')
                continue
            # If field is empty and not required let's ignore it
            if not field_value and not field.get('required'):
                continue
            field_serializer = cls.get_field_serializer(field_type)()
            try:
                field_serializer.to_internal_value(field_value)
                returned_properties[field_name] = field_value
            except (TypeError, ValueError) as err:
                errors[field_name] += err
            except ValidationError as err:
                errors[field_name] += err.detail

        if any([error for error in errors.values()]):
            raise ValidationError(errors)

        return returned_properties


class ContactSerializer(serializers.HyperlinkedModelSerializer):
    contact_list = serializers.HyperlinkedRelatedField(
        view_name='contacts:contactlist-detail',
        queryset=ContactList.objects.all())

    class Meta:
        model = Contact
        fields = [
            'url', 'contact_list', 'address', 'properties',
            'creation_date', 'update_date']
        read_only_fields = ['url', 'creation_date', 'update_date']
        list_serializer_class = ContactBulkSerializer
        extra_kwargs = {'url': {'view_name': 'contacts:contact-detail'}}

    def to_representation(self, obj):
        data = super().to_representation(obj)
        data['properties'] = PropertiesFieldHelper.to_representation(obj)
        return data

    def to_internal_value(self, data):
        obj = super().to_internal_value(data)
        try:
            obj['properties'] = PropertiesFieldHelper.to_internal_value(obj)
        except ValidationError as err:
            raise ValidationError({'properties': err.detail})

        return obj


class NestedContactSerializer(ContactSerializer):
    """ A ContactSerializer which list info comes from context (url)

    ... and not from input (api-submited body)
    """

    contact_list = serializers.HiddenField(
        default=serializers.CreateOnlyDefault(
            URLRelatedDefault('contact_list_pk', ContactList)))

    @classmethod
    def nest_properties(cls, data):
        """ Stick every unknown provided field into the "properties" hash

        >>> NestedContactSerializer.nest_properties(
        ... {'address': 'a@example.com','a':'b'}) == \
            {'properties': {'a': 'b'}, 'address': 'a@example.com'}
        True
        """
        nested_data = {'properties': {}}
        for k, v in data.items():
            if k in cls.Meta.fields:
                nested_data[k] = v
            else:
                sub_k = k.split('properties.')[-1]
                nested_data['properties'][sub_k] = v
        return nested_data

    def filter_properties(self, properties):
        # keep only the properties specified in fields url kwarg
        request = self.context['request']
        try:
            fields = request.query_params['fields']
        except KeyError:
            return properties
        else:
            return {k: v for k, v in properties.items() if k in fields}

    def to_internal_value(self, data):
        nested_data = self.nest_properties(data)
        nested_data['properties'] = self.filter_properties(
            nested_data['properties'])
        return super().to_internal_value(nested_data)


class ContactListListSerializer(serializers.Serializer):
    """ Used for list merge
    """
    contact_lists = ScopedHyperLinkedRelatedField(
        many=True,
        queryset=ContactList.objects.all(),  # FIXME,
        view_name='contacts:contactlist-detail',
        default=[])

    def __init__(self, master_list, *args, **kwargs):
        """
        :param master_list: the list that will receive others contacts
        """
        self.master_list = master_list
        super().__init__(*args, **kwargs)

    def validate_contact_lists(self, value):
        request = self.context['request']
        perm = ContactListMergePermission()

        # check that we can delete the lists we are going to empty
        # into the master one
        for contact_list in value:
            if not perm.can_user_do(request.user, contact_list, 'delete'):
                raise ValidationError(
                    "You don't have delete rights on {}".format(contact_list))
        return value

    def validate(self, data):
        if self.master_list in data['contact_lists']:
            raise ValidationError('Cannot merge a contact list with itself')
        return data



import collections

from rest_framework import viewsets
from rest_framework.reverse import reverse
from rest_framework.response import Response
from rest_framework.decorators import detail_route

from munch.apps.campaigns.models import Mail
from munch.apps.optouts.models import OptOut
from munch.apps.optouts.api.v1.filters import OptOutFilter
from munch.apps.optouts.api.v1.serializers import OptOutSerializer

from ...models import Category
from ...utils.views import filtered
from ...utils.views import paginated
from ...utils.views import MunchModelViewSetMixin
from ...utils.views import OrganizationOwnedViewSetMixin
from .filters import CategoryFilter
from .serializers import CategorySerializer
from .permissions import CategoryPermission


class CategoryViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = Category
    serializer_class = CategorySerializer
    filter_class = CategoryFilter
    permission_classes = [CategoryPermission]

    @detail_route(methods=['get'])
    def stats(self, request, pk, format=None):
        category = self.get_object()
        return Response(category.mk_stats())

    @detail_route(methods=['get'])
    def messages_stats(self, request, pk, format=None):
        # FIXME : might be optimized
        stats_dict = [
            collections.OrderedDict([
                ('url', reverse(
                    'campaigns:message-detail',
                    kwargs={'pk': m.pk}, request=request)),
                ('name', m.name),
                ('creation_date', m.creation_date),
                ('stats', m.mk_stats())
            ])
            for m in self.get_object().messages.all()
        ]
        return Response(stats_dict)

    @detail_route(methods=['get'])
    @paginated(OptOutSerializer)
    @filtered(OptOutFilter)
    def opt_outs(self, request, pk, format=None):
        identifiers = Mail.objects.filter(
            message__category=self.get_object()).values_list(
            'identifier', flat=True)
        return OptOut.objects.filter(identifier__in=identifiers)



def exception_handler(exc, context):
    # Replace the POST data of the Django request with the parsed
    # data from the DRF
    # Necessary because we cannot read request data/stream more than once
    # This will allow us to see the parsed POST params
    # in the rollbar exception log
    # Based on https://github.com/tomchristie/django-rest-framework/pull/1671
    from rest_framework import views
    from django.http.request import QueryDict
    query = QueryDict('', mutable=True)
    try:
        if not isinstance(context['request'].data, dict):
            query.update({'_bulk_data': context['request'].data})
        else:
            query.update(context['request'].data)
        context['request']._request.POST = query
    except:
        pass
    return views.exception_handler(exc, context)



import datetime
from calendar import timegm

from rest_framework_jwt.settings import api_settings


def jwt_payload_handler(user):
    """ Custom handler to not handle email and pk """
    payload = {
        'username': user.identifier,
        'exp': datetime.datetime.utcnow() + api_settings.JWT_EXPIRATION_DELTA}

    if api_settings.JWT_ALLOW_REFRESH:
        payload['orig_iat'] = timegm(datetime.datetime.utcnow().utctimetuple())

    return payload



import os


def configuration(f):
    import click
    from functools import update_wrapper

    @click.pass_context
    def inner(ctx, *args, **kwargs):
        # HACK: We can't call `configure()` from within tests
        # since we don't load config files from disk, so we
        # need a way to bypass this initialization step
        if os.environ.get('_MUNCH_SKIP_CONFIGURATION') != '1':
            from munch.runner import configure
            configure()
        return ctx.invoke(f, *args, **kwargs)
    return update_wrapper(inner, f)



import random
import logging
from unittest.mock import patch
from unittest.mock import MagicMock

import PIL.Image
from django.test import TestCase
from django.core.files.base import ContentFile
from django.core.files.temp import NamedTemporaryFile

from munch.apps.users.tests.factories import UserFactory

from .models import Image
from .models import UploadDuplicateError


logging.disable(logging.CRITICAL)


def mock_get_app_url(*args, **kwargs):
    def get_app_url():
        return 'http://uploads.example.com'

    m = MagicMock()
    m.side_effect = get_app_url()
    return m


class ImageTestCaseMixin:
    def create_random_image(self, width=800, height=600):
        color = (
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255))
        image = PIL.Image.new('RGB', (width, height), color)
        temp = NamedTemporaryFile(suffix='.png')
        image.save(temp, 'png')
        temp.seek(0)
        return temp


@patch('munch.core.mail.utils.get_app_url', mock_get_app_url)
class ImageTestCase(TestCase, ImageTestCaseMixin):
    def setUp(self):
        self.admin = UserFactory(
            is_admin=True, is_active=True, is_superuser=True)
        self.admin.organization.can_attach_files = False
        self.admin.organization.can_external_optout = False
        self.admin.organization.save()
        self.organization = self.admin.organization

    def test_image_creation(self):
        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image().read())

        image.file.save('random_image.png', file, save=False)
        image.save()
        self.assertNotEqual('', image.hash)

        image_file = PIL.Image.open(image.file.file)
        self.assertEqual((600, 450), image_file.size)

    def test_image_duplicate(self):
        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image().read())

        image.file.save('random_image.png', file, save=False)
        image.save()

        image = Image(organization=self.organization)

        image.file.save('random_image2.png', file, save=False)
        with self.assertRaises(UploadDuplicateError):
            image.save()

    def test_upload_differentImage(self):
        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image(100, 100).read())

        image.file.save('random_image.png', file, save=False)
        image.save()

        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image(101, 101).read())

        image.file.save('random_image.png', file, save=False)
        try:
            image.save()
        except UploadDuplicateError:
            self.fail('Creating a new image should not raise an exception.')

    def test_image_resizing(self):
        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image(100, 100).read())

        image.file.save('random_image.png', file, save=False)
        image.save()

        image_file = PIL.Image.open(image.file.file)
        self.assertEqual((100, 100), image_file.size)

        image = Image(organization=self.organization)
        file = ContentFile(self.create_random_image(100, 100).read())

        image.file.save('random_image.png', file, save=False)
        image.width = 50
        image.save()

        image_file = PIL.Image.open(image.file.file)
        self.assertEqual((50, 50), image_file.size)

    def test_backend(self):
        pass  # will be implemented when new backends will be available

    def test_resize_in_edit_mode(self):
        image = Image(organization=self.organization)
        image.resize_image = MagicMock()
        file = ContentFile(self.create_random_image(100, 100).read())

        image.file.save('random_image.png', file, save=False)
        image.save()

        image.resize_image.assert_called_with()

        image.resize_image.reset_mock()

        image.expiration = '10:00'
        image.save()

        self.assertFalse(
            image.resize_image.called,
            'Saving an existing image should not call '
            'resize_image, except if width has changed')

        image.resize_image.reset_mock()

        image.width = 200
        image.save()

        self.assertTrue(
            image.resize_image.called,
            'Saving an existing image with a new width should '
            'call resize_image')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-22 10:32
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.hstore
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import munch.core.utils.models
import munch.apps.transactional.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0002_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='Mail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient', models.EmailField(max_length=254, verbose_name='recipient')),
                ('identifier', models.CharField(db_index=True, default=munch.apps.transactional.models.get_mail_identifier, max_length=35, unique=True, verbose_name='identifier')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now, verbose_name='creation date')),
                ('first_status_date', models.DateTimeField(blank=True, null=True, verbose_name='first status date')),
                ('latest_status_date', models.DateTimeField(blank=True, null=True, verbose_name='latest status date')),
                ('delivery_duration', models.DurationField(null=True, verbose_name='delivery duration')),
                ('had_delay', models.BooleanField(default=False, verbose_name='had delay')),
                ('headers', django.contrib.postgres.fields.hstore.HStoreField(verbose_name='headers')),
                ('sender', models.EmailField(max_length=254, verbose_name='sender')),
                ('track_open', models.BooleanField(default=False, help_text='Detect who opens message by adding hidden image.', verbose_name='track opens')),
                ('track_clicks', models.BooleanField(default=False, help_text='Detect which links were clicked by replacing original links by redirections', verbose_name='track clicks')),
                ('msg_links', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='message links')),
                ('curstatus', models.CharField(choices=[('unknown', 'unknown'), ('ignored', 'ignored'), ('deleted', 'deleted'), ('queued', 'queued'), ('sending', 'sending'), ('delayed', 'delayed (retrying)'), ('delivered', 'delivered to remote MTA'), ('dropped', 'not delivered (soft-bounce)'), ('bounced', 'not delivered (hard-bounce)')], default='unknown', max_length=20, verbose_name='current status')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
            ],
            options={
                'abstract': False,
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='MailBatch',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='creation date')),
                ('identifier', models.CharField(db_index=True, default=munch.core.mail.utils.mk_base64_uuid, editable=False, max_length=35, unique=True, verbose_name='identifier')),
                ('name', models.CharField(max_length=60, verbose_name='name')),
                ('msg_links', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='message links')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='mailbatches', to='core.Category', verbose_name='category')),
            ],
            options={
                'verbose_name': 'mail batch',
                'verbose_name_plural': 'mail batches',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='MailStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('unknown', 'unknown'), ('ignored', 'ignored'), ('deleted', 'deleted'), ('queued', 'queued'), ('sending', 'sending'), ('delayed', 'delayed (retrying)'), ('delivered', 'delivered to remote MTA'), ('dropped', 'not delivered (soft-bounce)'), ('bounced', 'not delivered (hard-bounce)')], default='queued', max_length=15, verbose_name='status')),
                ('creation_date', models.DateTimeField(verbose_name='creation_date')),
                ('source_ip', models.GenericIPAddressField(blank=True, null=True, verbose_name='source ip')),
                ('source_hostname', models.CharField(blank=True, max_length=100, null=True, verbose_name='source hostname')),
                ('destination_domain', models.CharField(max_length=150, verbose_name='destination domain')),
                ('status_code', models.CharField(blank=True, help_text='X.XXX.XXX, as defined in RFC3463', max_length=25, validators=[django.core.validators.RegexValidator(message='status code should follow rfc3463', regex='([234](\\.\\d{1,3}){2})')], verbose_name='SMTP status code')),
                ('raw_msg', models.TextField(verbose_name='raw message')),
                ('mail', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='statuses', to='transactional.Mail', verbose_name='mail')),
            ],
            options={
                'get_latest_by': 'creation_date',
                'verbose_name': 'mail status',
                'verbose_name_plural': 'mail statuses',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.AddField(
            model_name='mail',
            name='batch',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='mails', to='transactional.MailBatch', verbose_name='batch'),
        ),
        migrations.AddField(
            model_name='mail',
            name='message',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='transactional_mail', to='core.RawMail', verbose_name='message'),
        ),
        migrations.AlterUniqueTogether(
            name='mailbatch',
            unique_together=set([('author', 'name')]),
        ),
    ]



from django.test import TestCase
from django.core import mail as django_mail
from django.test.utils import override_settings

from munch.apps.users.tests.factories import UserFactory
from munch.apps.optouts.tests.factories import OptOutFactory

from ..models import OptOut
from .factories import MailFactory
from .factories import MessageFactory


@override_settings(SKIP_SPAM_CHECK=True)
class NotificationTests(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.user.organization.contact_email = 'foo@bar'
        self.user.organization.settings.save()
        self.message = MessageFactory(author=self.user)
        django_mail.outbox = []

    def test_notify_optout_by_bounce(self):
        mail = MailFactory(message=self.message)
        mail.message.author.organization.settings.notify_optouts = True
        mail.message.author.organization.settings.save()
        OptOutFactory(
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_BOUNCE)

        self.assertEqual(len(django_mail.outbox), 0)

    def test_notify_optout(self):
        mail = MailFactory(message=self.message)
        mail.message.author.organization.settings.notify_optouts = True
        mail.message.author.organization.settings.save()
        OptOutFactory(
            author=mail.message.author,
            category=mail.message.category,
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_WEB)

        first_mail = django_mail.outbox[0]

        self.assertIn('no-reply', first_mail.from_email)

    def test_notify_sending(self):
        MailFactory(message=self.message)
        self.message.status = 'sending'
        self.message.save()  # Fires the message

        first_mail = django_mail.outbox[0]

        self.assertIn(
            self.user.organization.contact_email, first_mail.to)

    def test_notify_sending_good_count(self):
        MailFactory(message=self.message)
        MailFactory(message=self.message)

        self.message.status = 'sending'
        self.message.save()  # Fires the message

        first_mail = django_mail.outbox[0]

        self.assertIn('2 recipient(s) will receive', first_mail.body)
        self.assertIn("no recipients have been ignored", first_mail.body)

    def test_notify_sent(self):
        MailFactory(message=self.message)
        MailFactory(message=self.message)

        self.message.status = self.message.SENDING
        self.message.save()
        django_mail.outbox = []

        self.message.status = self.message.SENT
        self.message.save()
        first_mail = django_mail.outbox[0]
        self.assertIn(
            self.user.organization.contact_email, first_mail.to)
        self.assertIn('Total emails: 2', first_mail.body)



from rest_framework.reverse import reverse


def get_api_root_urls(request):
    return {
        'contacts': reverse('v1:contacts:contact-list', request=request),
        'contacts/queues': reverse(
            'v1:contacts:contactqueue-list', request=request),
        'contacts/policies': reverse(
            'v1:contacts:contactlistpolicy-list', request=request),
        'contacts/lists': reverse(
            'v1:contacts:contactlist-list', request=request)}



from django.conf import settings

from munch.core.utils import rgetattr
from munch.core.mail.utils.parsers import extract_domain


def add_google_fbl_header(headers, mail, filters_settings):
    if headers and mail and filters_settings:
        if extract_domain(mail.recipient) == 'gmail.com':
            gtags = []
            for s in filters_settings['GOOGLE_FBL_TAGS'][:3]:
                attr = rgetattr(mail, s, None)
                if attr:
                    gtags.append(str(attr))
            if gtags:
                val = '{}:{}'.format(
                    (':').join(gtags),
                    settings.GOOGLE_FBL_SENDER_ID
                )
            else:
                val = settings.GOOGLE_FBL_SENDER_ID
            headers.update({'Feedback-ID': val})



import csv
import six

from django.conf import settings
from rest_framework.parsers import BaseParser
from rest_framework.exceptions import ParseError
from rest_framework_csv.orderedrows import OrderedRows

"""
This module is mostly a copy of rest_framework_csv.parsers. This should be
monkeypatched instead.
"""


def preprocess_stream(stream, charset):
    if six.PY2:
        # csv.py doesn't do Unicode; encode temporarily:
        return (chunk.encode(charset) for chunk in stream)
    else:
        return stream


def postprocess_row(row, charset):
    if six.PY2:
        # decode back to Unicode, cell by cell:
        return [cell.decode(charset) for cell in row]
    else:
        return row


def unicode_csv_reader(csv_data, dialect=csv.excel, charset='utf-8', **kwargs):
    csv_data = preprocess_stream(csv_data, charset)
    csv_reader = csv.reader(csv_data, dialect=dialect, **kwargs)
    for row in csv_reader:
        yield postprocess_row(row, charset)


def universal_newlines(stream):
    for intermediate_line in stream:
        # It's possible that the stream was not opened in universal
        # newline mode. If not, we may have a single "row" that has a
        # bunch of carriage return (\r) characters that should act as
        # newlines. For that case, lets call splitlines on the row. If
        # it doesn't have any newlines, it will return a list of just
        # the row itself.
        for line in intermediate_line.splitlines():
            # this is a fix for issue #30
            # https://github.com/mjumbewu/django-rest-framework-csv/issues/30
            # causing CSV requests to crash
            yield line.decode()


class CSVParser(BaseParser):
    """
    Parses CSV serialized data.

    The parser assumes the first line contains the column names.
    """

    media_type = 'text/csv'

    def parse(self, stream, media_type=None, parser_context=None):
        parser_context = parser_context or {}
        delimiter = parser_context.get('delimiter', ',')

        try:
            encoding = parser_context.get('encoding', settings.DEFAULT_CHARSET)
            rows = unicode_csv_reader(universal_newlines(stream),
                                      delimiter=delimiter, charset=encoding)
            data = OrderedRows(next(rows))
            for row in rows:
                row_data = dict(zip(data.header, row))
                data.append(row_data)
            return data
        except Exception as exc:
            raise ParseError('CSV parse error - %s' % str(exc))



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class DomainsApp(AppConfig):
    name = 'munch.apps.domains'
    verbose_name = _('Domains')

    def ready(self):
        import munch.apps.domains.signals  # noqa
        import munch.apps.domains.api.v1.urls  # noqa



from django.conf import settings
import lxml
import lxml.html

from .models import HostedImage
from .models import InlineImage

from munch.apps.campaigns.exceptions import WrongHTML


def handle_images(html, detach_images=False, organization=None, **kwargs):
    """ Detach base64 images and others if detach_images is enabled
    """
    tree = lxml.html.fromstring(html)

    for img in tree.cssselect('img'):
        try:
            src = img.attrib['src']
        except KeyError:
            raise WrongHTML('<img> devrait avoir un attribut "src"')
        if src.startswith('data:image/'):
            # TODO: handle ValueError
            image = InlineImage(src, organization=organization)
            url = image.store()
            img.set('src', url)
        else:
            if detach_images and organization:
                image = HostedImage(src, organization=organization)
                url = image.store()
                img.set('src', url)
    return lxml.html.tostring(tree).decode()


def set_web_link_url(text, web_view_url=None, **kwargs):
    """
    :param text_input: can be any text, including html
    """
    placeholder = settings.HOSTED['WEB_LINK_PLACEHOLDER']
    if web_view_url:
        if placeholder in text:
            return text.replace(placeholder, web_view_url)
    return text



from django import template
from django.utils.translation import ugettext_lazy as _

register = template.Library()


@register.filter
def divide(value, arg):
    try:
        return int(value) / int(arg)
    except (ValueError, ZeroDivisionError):
        return None


@register.filter
def multiply(value, arg):
    try:
        return int(value) * int(arg)
    except (ValueError):
        return None


@register.filter
def percent(v):
    return '{:.0%}'.format(float(v))


@register.filter
def enabled_label(v):
    if v:
        return _('enabled')
    return _('disabled')



from django.test import TestCase

from munch.apps.users.tests.factories import UserFactory
from munch.apps.optouts.tests.factories import OptOutFactory

from .factories import MailFactory
from .factories import MessageFactory


class AllModelsTests(TestCase):
    def test_get_organization(self):
        """
        Check that all the category app define a get_organization() method
        """
        user = UserFactory()
        message = MessageFactory(author=user)
        mail = MailFactory(message=message)
        optout = OptOutFactory(
            author=user, category=message.category,
            identifier=mail.identifier, address=mail.recipient)
        klasses = [
            user.organization, message, mail.statuses.first(), mail, optout]

        for obj in klasses:
            self.assertEqual(obj.get_organization().pk, user.organization.pk)



import re
import email
import socket
from collections import OrderedDict
from functools import total_ordering

import dns.resolver
import dns.exception


class SpamCheckerError(Exception):
    def __init__(self, e):
        """
        @param e : the underlying exception
        """
        self.sub_e = e

    def __str__(self):
        return str(self.sub_e)


class SpamChecker:
    def __init__(self, service_name=None, host=None, port=None):
        self.service_name = service_name

        if host and port:
            self.host = host
            self.port = int(port)
        else:
            # first, resolve the host/port from service name
            try:
                results = dns.resolver.query(service_name, 'SRV')
                prio, weight, port, target = results[0].to_text().split(' ')
            except dns.exception.DNSException as e:
                raise SpamCheckerError(e)

            self.port = int(port)
            self.host = target

    def check(self, msg):
        """
        @param msg : a string containing the message to check
        @return status, score, message
        """
        msg_bytes = msg.encode()
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(2)
            s.connect((self.host, self.port))
            s.sendall(b'PROCESS SPAMC/1.4')
            s.sendall(b"\r\n")
            s.sendall(("Content-length: %s" % len(msg_bytes)).encode())
            s.sendall(b"\r\n")
            s.sendall(b"\r\n")
            s.sendall(msg_bytes)
            s.shutdown(1)
            socketfile = s.makefile("rb")
            line1_info = socketfile.readline()
            socketfile.readline()
            socketfile.readline()
            socketfile.readline()
            content = socketfile.read()
            if len(line1_info) == 0:
                raise SpamCheckerError("Spamd response is empty")

            answer = line1_info.strip().split()
            if len(answer) != 3:
                raise SpamCheckerError("Invalid Spamd answer: {}".format(
                    answer))
            (version, number, status) = answer
            if status.decode() != 'EX_OK':
                raise SpamCheckerError('Invalid Spamd status: {}'.format(
                    status))

        except (socket.error) as e:
            raise SpamCheckerError(e) from None

        else:
            # parse mail
            return SpamResult.from_headers(email.message_from_bytes(content))


class SpamResult:
    SPAM_STATUS_HEADER = 'X-Spam-Status'
    SPAM_REPORT_HEADER = 'X-Spam-Report'

    r_spam_status = re.compile(
        r'\s*(?P<is_spam>Yes|No),.*\s+score=(?P<score>\-?\d+\.\d+) .*',
        re.MULTILINE | re.DOTALL)
    r_spam_report = re.compile((
        r'\s+\*\s+(?P<score>\d+\.\d+)\s+(?P<name>[A-Z0-9_]+)'
        r'\s+(?P<main_content>.*)'
        r'(\n\s+\*      (?P<extra_content>.*))*'), re.MULTILINE)

    def __init__(self, score, is_spam, checks, error=None):
        self.score = float(score)
        self.is_spam = is_spam
        self._checks = checks
        self.error = error

    @classmethod
    def from_headers(cls, headers):
        for i in (cls.SPAM_STATUS_HEADER, cls.SPAM_REPORT_HEADER):
            if i not in headers:
                raise SpamCheckerError('Header {} not present'.format(i))
        status = headers[cls.SPAM_STATUS_HEADER]
        score = cls.r_spam_status.match(status).group('score')
        is_spam = cls.r_spam_status.match(status).group('is_spam') == 'Yes'

        checks = []
        for m in cls.r_spam_report.finditer(headers[cls.SPAM_REPORT_HEADER]):
            name = m.group('name')
            check_score = m.group('score')
            description = m.group('main_content')
            if m.group('extra_content'):
                description += ' ' + m.group('extra_content')
            checks.append(SpamCheckResult(check_score, name, description))
        return cls(score, is_spam, checks)

    def serialize(self):
        return OrderedDict((
            ('error', self.error),
            ('is_spam', None if self.error else self.is_spam),
            ('score', None if self.error else self.score),
            ('checks',
             [i.serialize() for i in sorted(self._checks, reverse=True)])))

    @property
    def checks(self):
        return [i.serialize() for i in sorted(self._checks, reverse=True)]


@total_ordering
class SpamCheckResult:
    def __init__(self, score, name, description):
        self.score = float(score)
        self.name = name
        self.description = description

    def __str__(self):
        return '{}: {}'.format(self.name, self.score)

    def serialize(self):
        return {
            'name': self.name,
            'score': self.score,
            'description': self.description
        }

    def __lt__(self, other):
        return self.score < other.score



from django.utils.translation import ugettext as _

from rest_framework.exceptions import APIException


class UserDisabledException(APIException):
    status_code = 403
    default_detail = _('User account is disabled')



from django.conf import settings
from django.core.exceptions import ValidationError
from django.contrib.auth.password_validation import MinimumLengthValidator as DjangoMinimumLengthValidator  # noqa
from django.utils.translation import ungettext


class MinimumLengthValidator(DjangoMinimumLengthValidator):
    def __init__(self):
        super().__init__(getattr(settings, 'PASSWORD_MIN_LENGTH', 6))


class MinimumDigitValidator(object):
    def __init__(self):
        self.min_digit = getattr(settings, 'PASSWORD_MIN_DIGIT', 0)

    def validate(self, password, user=None):
        if self.min_digit:
            num_digit = sum(c.isdigit() for c in password)
            if num_digit < self.min_digit:
                raise ValidationError(
                    ungettext(
                        'This password must contain at least %(mini)d digit',
                        'This password must contain at least %(mini)d digits',
                        self.min_digit) % {'mini': self.min_digit})

    def get_help_text(self):
        help_text = ''
        if self.min_digit:
            help_text = ungettext(
                'Your password must contain at least %(mini)d digit',
                'Your password must contain at least %(mini)d digits',
                self.min_digit) % {'mini': self.min_digit}
        return help_text


class MinimumUppercaseValidator(object):
    def __init__(self):
        self.min_upper = getattr(settings, 'PASSWORD_MIN_UPPER', 0)

    def validate(self, password, user=None):
        if self.min_upper:
            num_upper = sum(c.isupper() for c in password)
            if num_upper < self.min_upper:
                raise ValidationError(
                    ungettext(
                        'This password must contain at least %(mini)d uppercase character',  # noqa
                        'This password must contain at least %(mini)d uppercase characters',  # noqa
                        self.min_upper) % {'mini': self.min_upper})

    def get_help_text(self):
        help_text = ''
        if self.min_upper:
            help_text = ungettext(
                'Your password must contain at least %(mini)d uppercase character',  # noqa
                'Your password must contain at least %(mini)d uppercase characters',  # noqa
                self.min_upper) % {'mini': self.min_upper}
        return help_text



from django.conf.urls import url

from . import views

urlpatterns = [
    url(
        r'^subscriptions/(?P<identifier>[\w\-_]+)/optout/$',
        views.unsubscribe, name='unsubscribe'),
    url(
        r'^subscriptions/(?P<identifier>[\w\-_]+)/unsubscribed/$',
        views.unsubscribed, name='unsubscribed')
]



from datetime import timedelta

from django.test import TestCase
from django.utils import timezone
from django.test.utils import override_settings

from munch.apps.users.tests.factories import UserFactory

from ..models import MailStatus

from .factories import MailFactory
from .factories import MessageFactory
from .factories import MailStatusFactory


class MailStatusManagerTest(TestCase):
    def test_duration_zeroitem(self):
        self.assertEqual(MailStatus.objects.none().duration(), timedelta())


class MailStatusTest(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)
        self.mail = MailFactory(message=self.message)
        MailStatusFactory(mail=self.mail, status=MailStatus.QUEUED)
        MailStatusFactory(mail=self.mail, status=MailStatus.SENDING)

    def test_match_policy_ok(self):
        mail_status = MailStatusFactory(
            mail=self.mail, status=MailStatus.BOUNCED, status_code='5.1.1')
        ok_policy = (['6.', '5.'], 3, 1)
        self.assertTrue(mail_status.match_policy(ok_policy))

    def test_match_policy_empty(self):
        mail_status = MailStatusFactory(
            mail=self.mail, status=MailStatus.BOUNCED,
            creation_date=timezone.now(), status_code='5.1.1')
        ok_policy = ([''], 3, 1)
        self.assertTrue(mail_status.match_policy(ok_policy))

    def test_nomatch_policy_code(self):
        mail_status = MailStatusFactory(
            mail=self.mail, status=MailStatus.BOUNCED,
            creation_date=timezone.now(), status_code='5.1.1')
        othercode_policy = (['3.'], 3, 1)
        self.assertFalse(mail_status.match_policy(othercode_policy))

    def test_nomatch_policy_date(self):
        mail_status = MailStatusFactory(
            mail=self.mail, status=MailStatus.BOUNCED,
            creation_date=timezone.now() - timedelta(days=2),
            status_code='5.1.1')
        tooold_policy = (['5.'], 3, 1)
        self.assertFalse(mail_status.match_policy(tooold_policy))

    @override_settings(BOUNCE_POLICY=[(['5.'], 3, 1)])
    def test_should_optout(self):
        user = UserFactory()

        def mk_message(i):
            message = MessageFactory(author=user)
            mail = MailFactory(recipient='i@example.com', message=message)

            MailStatusFactory(
                status=MailStatus.QUEUED, raw_msg='queued',
                mail=mail, status_code='5.1.1')
            MailStatusFactory(
                status=MailStatus.SENDING, raw_msg='sent',
                mail=mail, status_code='5.1.1')
            return MailStatusFactory(
                status=MailStatus.BOUNCED, raw_msg='Hardbounced',
                mail=mail, status_code='5.1.1')

        b1 = mk_message(1)
        self.assertFalse(b1.should_optout())
        b2 = mk_message(2)
        self.assertFalse(b2.should_optout())
        b3 = mk_message(3)
        self.assertTrue(b3.should_optout())



from django.contrib import admin

from .tasks import validate_sending_domain_field
from .models import SendingDomain


class SendingDomainAdmin(admin.ModelAdmin):
    readonly_fields = [
        'dkim_status', 'dkim_status_date', 'creation_date',
        'update_date', 'app_domain_status', 'app_domain_status_date']
    actions = ['revalidate']
    list_display = [
        'name', 'organization', 'creation_date',
        'update_date', 'dkim_status', 'dkim_status_date',
        'app_domain_status', 'app_domain_status_date']
    search_fields = ['name', 'app_domain']
    list_filter = [
        'creation_date', 'update_date', 'dkim_status', 'dkim_status_date',
        'app_domain_status', 'app_domain_status_date']
    filter_horizontal = ('alt_organizations',)

    def revalidate(self, request, queryset):
        for domain in queryset.only('id'):
            validate_sending_domain_field.apply_async([domain.id, 'dkim'])
            validate_sending_domain_field.apply_async(
                [domain.id, 'app_domain'])


class SendingDomainInline(admin.TabularInline):
    model = SendingDomain
    extra = 0
    ordering = ['creation_date']
    readonly_fields = [
        'creation_date', 'update_date', 'dkim_status', 'dkim_status_date']


admin.site.register(SendingDomain, SendingDomainAdmin)



from django.utils.module_loading import import_string


class MuncherRunner:
    def __init__(self, settings, filters_settings_key):
        self.settings = settings
        self.filters_settings_key = filters_settings_key

    def process(self, *args, **kwargs):
        raise NotImplementedError


class ContentMuncherRunner(MuncherRunner):
    def process(self, indata, *args, **kwargs):
        registered_functions = self.settings.get(
            self.filters_settings_key, [])

        outdata = indata
        for func_dotted_path in registered_functions:
            func = import_string(func_dotted_path)
            outdata = func(outdata, *args, **kwargs)

        return outdata


class HeadersMuncherRunner(MuncherRunner):
    def process(self, *args, **kwargs):
        registered_functions = self.settings.get(
            self.filters_settings_key, [])

        for func_dotted_path in registered_functions:
            func = import_string(func_dotted_path)
            func(*args, **kwargs)



from rest_framework import serializers
from rest_framework.views import APIView
from rest_framework.generics import get_object_or_404
from rest_framework.serializers import ValidationError as DRFValidationError

from .api import get_subroute_method
from .permissions import MunchResourcePermission
from .pagination import CountedPageNumberPagination


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class OrganizationOwnedViewSetMixin:
    """ Auto-handle a "organization" field according to the current User

    Supposed to be used with utils.OwnedModelMixin models.
    """

    @classmethod
    def mk_permission_classes(cls, extra_perms=[]):
        """ Dynamically build permission classes

        based on underlying model
        """
        return [MunchResourcePermission.mk_class(cls.model)] + extra_perms

    @classproperty
    @classmethod
    def permission_classes(cls):
        return cls.mk_permission_classes()

    @classmethod
    def _get_owner_permission(cls):
        for i in cls.permission_classes:
            if issubclass(i, MunchResourcePermission):
                return i()

        raise ValueError(
            'That view should use a MunchResourcePermission class')

    def get_object(self):
        """ That override is to get 403 instead of 404 on single object ops

        Mostly a copy-paste from GenericAPIView
        """
        queryset = self.filter_queryset(self.model.objects.all())
        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, lookup_url_kwarg))
        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        obj = get_object_or_404(queryset, **filter_kwargs)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)
        return obj

    def get_queryset(self):
        perm = self._get_owner_permission()
        qs = perm.filter_queryset(
            self.model.objects.all(), self.request)

        if hasattr(self, 'ordering'):
            return qs.order_by(self.ordering)
        return qs

    def _auto_set_owner_author(self):
        kwargs = {}
        if self.model.has_owner_direct_link():
            kwargs[self.model.owner_path] = self.request.user.organization
        if self.model.has_author_direct_link() and self.model.defines_author():
            kwargs[self.model.author_path] = self.request.user
        return kwargs

    def _check_author_link_permissions(self, serializer_row):
        """ Checks if legitimate to link to a resource

        We consider that legitimacy comes from having right to edit the
        resource we link to.
        """
        if (not self.model.has_author_direct_link() and
                not self.model.is_author_class()):
            referent_obj = self.model.get_directly_linked_obj(serializer_row)
            perm = self._get_owner_permission()
            if not perm.can_user_do(self.request.user, referent_obj, 'change'):
                raise DRFValidationError(
                    'You tried to link to a forbidden resource : {}'.format(
                        referent_obj))

    def perform_create(self, serializer):
        kwargs = self._auto_set_owner_author()
        if getattr(serializer, 'many', False):
            serializer_data_rows = serializer.validated_data
        else:
            serializer_data_rows = [serializer.validated_data]

        for row in serializer_data_rows:
            self._check_author_link_permissions(row)

        serializer.save(**kwargs)

    def perform_update(self, serializer):
        # treatement is strictly similar
        return self.perform_create(serializer)


class OrganizationOptionnallyOwnedViewSetMixin(OrganizationOwnedViewSetMixin):
    pass


class MunchModelViewSetMixin:
    def get_view_description(self, html=False):
        """
        Return some descriptive text for the view, as used in OPTIONS responses
        and in the browsable API.
        """
        func = self.settings.VIEW_DESCRIPTION_FUNCTION
        return func(self.__class__, html, instance=self)

    def filter_queryset(self, queryset):
        """
        If we are in a list_route or detail_route, just pass through
        """
        if not get_subroute_method(self):
            for backend in list(self.filter_backends):
                queryset = backend().filter_queryset(
                    self.request, queryset, self)
        return queryset


class NestedViewMixin:
    def get_parent_object(self, pk):
        parent_obj = get_object_or_404(self.parent_model, pk=pk)
        self.check_object_permissions(self.request, parent_obj)
        return parent_obj

    def get_serializer(self, *args, **kwargs):
        """ Directly inspired from DRF GenericAPIView
        """
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        return self.serializer_class

    def get_serializer_context(self):
        return {'request': self.request}


class NestedView(NestedViewMixin, APIView):
    """ View meant to be nested within a resource

    It uses explicit url conf rather than router

    must define a `parent_model`
    may define a `serializer_class`

    Checks permissions on the parent object
    """
    serializer_class = serializers.Serializer


def paginated_response(queryset, serializer_class, request):
    """
    Pagination helper for a @detail route.
    http://stackoverflow.com/a/29144786/1377500

    :type serializer_class: rest_framework.serializers.Serializer
    :rtype:                 rest_framework.response.Response
    """
    paginator = CountedPageNumberPagination()
    page = paginator.paginate_queryset(queryset, request)
    serializer = serializer_class(
        page, many=True, context={'request': request})
    return paginator.get_paginated_response(serializer.data)


def filtered_qs(qs, filter_class, request):
    """ Apply a django-filter to a queryset against a request

    :type qs:           django.db.models.QuerySet
    :type filter_class: django_filters.FilterSet
    :return:            filtered queryset
    """
    return filter_class(request.query_params, queryset=qs).qs


def paginated(serializer_class):
    """ Decorator to transform a queryset return into a Paginated response.
    """
    def wrap(f):
        def wrapped_f(self, request, *args, **kwargs):
            qs = f(self, request, *args, **kwargs)
            return paginated_response(qs, serializer_class, request)
        # Preserve the attributes of original function
        for i in dir(f):
            if not i.startswith('_'):
                setattr(wrapped_f, i, getattr(f, i))
        return wrapped_f
    return wrap


def filtered(filter_class):
    """ Decorator to filter a queryset, keeping a queryset as return val
    """
    def wrap(f):
        def wrapped_f(self, request, *args, **kwargs):
            qs = f(self, request, *args, **kwargs)
            return filtered_qs(qs, filter_class, request)
        wrapped_f.filter_class = filter_class
        return wrapped_f
    return wrap



import logging

from django.dispatch import receiver
from django.db.models.signals import pre_save
from django.db.models.signals import post_save
from django.db.models.signals import post_delete

from munch.core.mail import backend

from .models import MailStatus
from .models import MessageAttachment

log = logging.getLogger(__name__)

# MailStatus
pre_save.connect(backend.pre_save_mailstatus_signal, sender=MailStatus)
post_save.connect(backend.post_save_mailstatus_signal, sender=MailStatus)


@receiver(post_delete, sender=MessageAttachment)
def auto_delete_message_attachment_on_delete(sender, instance, **kwargs):
    if not instance.file:
        return

    if MessageAttachment.objects.filter(
            message__author__organization=instance.message.author.organization,
            file=instance.file.name).only('id').exists():
        log.debug('Remaining message attachments ({})...'.format(
            instance.file.name))
        return

    instance.file.delete(save=False)



import os
from io import BytesIO
from hashlib import sha1
from datetime import timedelta

import PIL.Image
from django.db import models
from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.validators import MaxValueValidator
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.utils.http import urlsafe_base64_encode
from django.utils.module_loading import import_string
from django.contrib.postgres.fields import HStoreField

from munch.core.mail.utils import mk_base64_uuid
from munch.core.utils.models import AbstractOwnedModel

from munch.apps.users.models import Organization


def get_storage():
    backend = import_string(settings.UPLOAD_STORE['BACKEND'])
    return backend(base_url=settings.UPLOAD_STORE['URL'])


def upload_to(instance, filename):
    prefix = settings.UPLOAD_STORE['URL_PREFIX'].strip('/')
    content_type = '{}s'.format(instance.__class__.__name__.lower())
    return os.path.join(
        prefix, content_type, str(instance.organization.id), filename)


class AbstractFile(AbstractOwnedModel):
    FILE = 'file'
    IMAGE = 'image'
    KIND_CHOICES = (
        (FILE, _('File')),
        (IMAGE, _('Image')))

    identifier = models.CharField(
        max_length=35, db_index=True, unique=True,
        default=mk_base64_uuid, verbose_name=_('identifier'))
    file = models.FileField(upload_to=upload_to)
    kind = models.CharField(
        max_length=10, choices=KIND_CHOICES, default=FILE,
        verbose_name=_('type'))
    original_name = models.CharField(
        max_length=100, null=True, blank=True, verbose_name=_('original name'))
    creation_date = models.DateTimeField(
        _('creation date'), default=timezone.now)
    update_date = models.DateTimeField(_('update date'), blank=True)
    organization = models.ForeignKey(
        Organization, related_name='%(class)ss',
        verbose_name=_('organization'))
    details = HStoreField(null=True, blank=True, verbose_name=_('details'))

    owner_path = 'organization'
    author_path = AbstractOwnedModel.IRRELEVANT

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Leaving it here for tests. Dunno why...
        storage = get_storage()
        self._meta.get_field('file').storage = storage
        self.file.storage = storage

    def make_hash(self, content):
        _hash = urlsafe_base64_encode(sha1(content.read()).digest()).decode()
        content.seek(0)
        return _hash

    def save(self, *args, **kwargs):
        self.update_date = timezone.now()

        if not self.original_name:
            self.original_name = self.file.name[:100]

        if self.kind == self.IMAGE:
            content = self.pre_save_kind_image()
        else:
            content = self.pre_save_kind_file()

        # Compute a filename bashed on SHA1 hash
        # to avoid duplicates
        _hash = self.make_hash(content)
        extension = os.path.splitext(self.file.name)[1]
        file_name = '{}{}'.format(_hash, extension)
        # Save the file kind so that storage backend can
        # access it if needed
        content.kind = self.kind
        # Save FieldFile with new name and possibly altered content
        self.file.save(name=file_name, content=content, save=False)

        super().save(*args, **kwargs)

    ########
    # File #
    ########

    def pre_save_kind_file(self):
        # Nothing to do, simply return unaltered content
        return self.file

    #########
    # Image #
    #########

    def pre_save_kind_image(self):
        return self.resize_image()

    def resize_image(self):
        # original code for this method came from
        # http://snipt.net/danfreak/generate-thumbnails-in-django-with-pil/

        # If there is no image associated with this.
        # do not create thumbnail
        if not self.file:
            return

        content = self.file
        max_width = settings.UPLOAD_STORE['IMAGE_MAX_WIDTH']
        if self.details:
            requested_width = self.details.get('width')
            try:
                requested_width = int(requested_width)
            except (TypeError, ValueError):
                requested_width = int(max_width)
        else:
            requested_width = int(max_width)

        # We won't resize past our max width
        resize_width = min(requested_width, max_width)
        # Set our max thumbnail size in a tuple (max width, max height)
        NEW_SIZE = (requested_width, 9999)

        # Open original photo which we want to thumbnail using PIL's Image
        image = PIL.Image.open(self.file.file)

        if resize_width > image.size[0]:
            # We don't want to upscale images
            pass
        else:
            # Convert to RGB if necessary
            # Thanks to Limodou on DjangoSnippets.org
            # http://www.djangosnippets.org/snippets/20/
            #
            # I commented this part since it messes up my png files
            #
            # if image.mode not in ('L', 'RGB'):
            #     image = image.convert('RGB')

            # We use our PIL Image object to create the thumbnail,
            # which already has a thumbnail() convenience method that
            # contrains proportions.
            # Additionally, we use Image.ANTIALIAS to make the image look
            # better. Without antialiasing the image pattern
            # artifacts may result.
            image.thumbnail(NEW_SIZE, PIL.Image.ANTIALIAS)

            # Save the resized image
            temp_handle = BytesIO()
            image.save(temp_handle, image.format, quality=90)
            temp_handle.seek(0)

            # Create a SimpleUploadedFile from the image
            # which can be saved into a FieldFile
            content = SimpleUploadedFile(
                self.file.name, temp_handle.read(),
                content_type=PIL.Image.MIME[image.format])

        self.details['width'] = str(image.size[0])
        self.details['height'] = str(image.size[1])

        return content


# class Image(AbstractFile):
#     def save(self, *args, **kwargs):
#         self.kind = self.IMAGE
#         super().save(*args, **kwargs)


###############
# Legacy code #
###############
class UploadDuplicateError(Exception):
    def __init__(self, message, instance, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.instance = instance


class Upload(models.Model):
    upload_date = models.DateTimeField(auto_now=True)
    expiration = models.DurationField()
    organization = models.ForeignKey(Organization, related_name='%(class)ss')

    class Meta:
        abstract = True

    def __str__(self):
        return self.file.name

    def get_absolute_url(self):
        return self.file.url

    @staticmethod
    def get_storage():
        backend = import_string(settings.UPLOAD_STORE['BACKEND'])
        return backend(base_url=settings.UPLOAD_STORE['URL'])

    @staticmethod
    def get_path(instance, filename):
        prefix = settings.UPLOAD_STORE['URL_PREFIX'].strip('/')
        content_type = '{}s'.format(instance.__class__.__name__.lower())
        return os.path.join(
            prefix, content_type, str(instance.organization.id), filename)

    def get_hash(self, fd):
        content = fd.read()
        # append organization ID to generate hash scoped per organization
        content += str(self.organization.id).encode()
        _hash = urlsafe_base64_encode(sha1(content).digest()).decode()
        fd.seek(0)
        return _hash


class Image(Upload):
    hash = models.CharField(max_length=64, primary_key=True)
    file = models.ImageField(upload_to=Upload.get_path)
    width = models.PositiveSmallIntegerField(
        default=600, validators=[MaxValueValidator(600)])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Leaving it here for tests. Dunno why...
        self._meta.get_field('file').storage = Upload.get_storage()

        self.file.storage = Upload.get_storage()
        self.initial_width = self.width

    def save(self, *args, **kwargs):
        if not self.expiration:
            self.expiration = timedelta(0)
        if self._state.adding or self.width != self.initial_width:
            self.resize_image()
            if self._state.adding:
                if Image.objects.filter(pk=self.hash).exists():
                    raise UploadDuplicateError(
                        'This image already exists', instance=self)
        super().save(*args, **kwargs)

    def resize_image(self, save=True):
        # original code for this method came from
        # http://snipt.net/danfreak/generate-thumbnails-in-django-with-pil/

        # If there is no image associated with this.
        # do not create thumbnail
        if not self.file:
            return
        requested_width = self.width
        # Set our max thumbnail size in a tuple (max width, max height)
        NEW_SIZE = (self.width, 9999)

        # Open original photo which we want to thumbnail using PIL's Image
        image = PIL.Image.open(self.file.file)

        if requested_width > image.size[0]:
            # We don't want to upscale images, record a lower width instead
            self.width = image.size[0]
        else:
            # Convert to RGB if necessary
            # Thanks to Limodou on DjangoSnippets.org
            # http://www.djangosnippets.org/snippets/20/
            #
            # I commented this part since it messes up my png files
            #
            # if image.mode not in ('L', 'RGB'):
            #     image = image.convert('RGB')

            # We use our PIL Image object to create the thumbnail,
            # which already has a thumbnail() convenience method that
            # contrains proportions.
            # Additionally, we use Image.ANTIALIAS to make the image look
            # better. Without antialiasing the image pattern
            # artifacts may result.
            image.thumbnail(NEW_SIZE, PIL.Image.ANTIALIAS)

        # Save the resized image (or original)
        temp_handle = BytesIO()
        image.save(temp_handle, image.format, quality=90)
        temp_handle.seek(0)

        # Save image to a SimpleUploadedFile which can be saved into
        # ImageField
        suf = SimpleUploadedFile(
            os.path.split(self.file.name)[-1], temp_handle.read(),
            content_type=PIL.Image.MIME[image.format])
        # Save SimpleUploadedFile into image field
        extension = suf.name.rsplit('.', 1)[-1]
        self.hash = self.get_hash(suf)
        self.file.save(
            name='{}.{}'.format(self.hash, extension), content=suf, save=False)



from munch.apps.campaigns.exceptions import InvalidSubmitedData


class TooBigMedia(InvalidSubmitedData):
    def __init__(self, name, max_size):
        self.name = name
        self.max_size = max_size
        super().__init__()

    def __str__(self):
        return 'Le fichier "{}" est trop gros (> {:.2}Mio)'.format(
            self.name, self.max_size / 1024 / 1024)


class InvalidMimeType(InvalidSubmitedData):
    def __init__(self, mime_type):
        self.mimetype = mime_type
        super().__init__()

    def __str__(self):
        return '"{}" n\'est pas un type mime supporté'.format(self.mimetype)



import logging
from celery.exceptions import Reject

log = logging.getLogger(__name__)


class SoftFailure(Exception):
    pass


class HardFailure(Exception):
    pass


class RejectMissingField(Reject):
    def __init__(self, field, recipient, headers=None, *args, **kwargs):
        msg = 'Missing field {} in received mail'.format(field)
        if headers:
            log.warning(msg + str(headers))
        else:
            log.warning(msg)

        super().__init__(msg, *args, **kwargs)


class RejectUnknownIdentifier(Reject):
    def __init__(self, identifier, headers=None, *args, **kwargs):
        msg = 'No mail known with Message-Id {}'.format(identifier)
        if headers:
            log.warning(msg + ". Headers: \n{}".format(str(headers)))
        else:
            log.warning(msg)

        super().__init__(msg, *args, **kwargs)


class RejectInvalidDSN(Reject):
    def __init__(self, exc, headers=None, *args, **kwargs):
        if headers:
            log.warning(str(exc) + str(headers))
        else:
            log.warning(str(exc))

        super().__init__(str(exc), *args, **kwargs)


class RejectForbiddenTransition(Reject):
    def __init__(self, obj, exc, *args, **kwargs):
        super().__init__(
            'Forbidden transition on {} : {}'.format(obj, str(exc)),
            *args, **kwargs)



import re
from datetime import timedelta

from django.conf import settings
from django.utils import timezone
from celery.decorators import periodic_task
from celery.utils.log import get_task_logger

from .models import Contact
from .models import AbstractContact
from .models import CollectedContact

from munch.core.utils.tasks import AutoRetryTask


log = get_task_logger(__name__)


@periodic_task(run_every=timedelta(minutes=1))
def handle_opt_ins_expirations():
    """
    Set contacts and collected contacts to Expired status when expiration
    time passed without being validated (status: OK).
    Both contacts and collected contacts are concerned.
    """
    doi_expiration = timezone.now() - settings.CONTACTS['EXPIRATIONS'][
        'contact_queues:double-opt-in']
    contacts = CollectedContact.objects.filter(
        status=AbstractContact.PENDING,
        update_date__lt=doi_expiration,
        contact_queue__policies__name='DoubleOptIn')
    log.info(
        'Updating {} collected contact(s) to Expired'.format(contacts.count()))
    contacts.update(status=AbstractContact.EXPIRED)

    doi_expiration = timezone.now() - settings.CONTACTS['EXPIRATIONS'][
        'contact_lists:double-opt-in']
    contacts = Contact.objects.filter(
        status=AbstractContact.PENDING,
        update_date__lt=doi_expiration,
        contact_list__policies__name='DoubleOptIn')
    log.info('Updating {} contact(s) to Expired'.format(contacts.count()))
    contacts.update(status=AbstractContact.EXPIRED)


@periodic_task(run_every=timedelta(minutes=1))
def handle_bounce_expirations():
    """
    Set contacts and collected contacts to OK status when expiration time
    passed without being bounced.
    Both contacts and collected contacts are concerned.
    """
    bc_expiration = timezone.now() - settings.CONTACTS['EXPIRATIONS'][
        'contact_queues:bounce-check']
    contacts = CollectedContact.objects.exclude(
        contact_queue__policies__name='DoubleOptIn').filter(
            status=AbstractContact.PENDING,
            update_date__lt=bc_expiration,
            contact_queue__policies__name='BounceCheck')
    # We don’t want Double-Opt-In contacts here, because the “OK” status should
    # only be given by the confirmation link
    log.info('Accepting {} collected contact(s) that didn’t bounce'
             .format(contacts.count()))
    contacts.update(status=AbstractContact.OK)

    bc_expiration = timezone.now() - settings.CONTACTS['EXPIRATIONS'][
        'contact_lists:bounce-check']
    contacts = Contact.objects.exclude(
        contact_list__policies__name='DoubleOptIn').filter(
            status=AbstractContact.PENDING,
            update_date__lt=bc_expiration,
            contact_list__policies__name='BounceCheck')
    # We don’t want Double-Opt-In contacts here, because the “OK” status should
    # only be given by the confirmation link
    log.info('Accepting {} contact(s) that didn’t bounce'
             .format(contacts.count()))
    contacts.update(status=AbstractContact.OK)


@periodic_task(run_every=timedelta(minutes=1))
def handle_consumed_contacts_expirations():
    """
    Delete consumed contacts after an expiration time.
    Only collected contacts can be consumed.
    """
    filters = {
        'status': AbstractContact.CONSUMED,
        'update_date__lt': timezone.now() - settings.CONTACTS['EXPIRATIONS'][
            'contact_queues:consumed_lifetime']}
    contacts = CollectedContact.objects.filter(**filters)
    log.info('Deleting {} collected contact(s) that have been consumed'.format(
        contacts.count()))
    contacts.delete()


@periodic_task(run_every=timedelta(minutes=1))
def handle_failed_expirations():
    """
    Delete expired and bounced contacts after an expiration time.
    Both contacts and collected contacts can have these statuses.
    """
    filters = {
        'status__in': (AbstractContact.BOUNCED, AbstractContact.EXPIRED),
        'update_date__lt': None}

    filters.update({
        'update_date__lt': timezone.now() - settings.CONTACTS['EXPIRATIONS'][
            'contact_queues:failed_lifetime']})
    contacts = CollectedContact.objects.filter(**filters)
    log.info('Deleting {} collected contact(s) that bounced or expired'.format(
        contacts.count()))
    contacts.delete()

    filters.update({
        'update_date__lt': timezone.now() - settings.CONTACTS['EXPIRATIONS'][
            'contact_lists:failed_lifetime']})
    contacts = Contact.objects.filter(**filters)
    log.info('Deleting {} contact(s) that bounced or expired'.format(
        contacts.count()))
    contacts.delete()


@periodic_task(run_every=timedelta(minutes=1))
class handle_bounce(AutoRetryTask):
    def safe_run(self, msg):
        regexp = re.compile(r'^subscription-bounce\+(?P<uuid>[^@]+)@')
        match = regexp.match(msg['To'])
        if match:
            uuid = match.groups()[0]
            contact = None
            filters = {'uuid': uuid, 'status': AbstractContact.PENDING}
            try:
                contact = CollectedContact.objects.get(**filters)
            except CollectedContact.DoesNotExist:
                try:
                    contact = Contact.objects.get(**filters)
                except Contact.DoesNotExist:
                    pass

            if not contact:
                return

            contact.status = CollectedContact.BOUNCED
            contact.save()
            log.info(
                'Contact subscription {} is bounced'.format(contact.address))



import datetime

from django.db import models
from django.db.models import Q
from django.db.models import Sum


class MedianQuerySetMixin:
    def median(self, field, order=True):
        """ Returns the median field value of a given field

        :param order: should we order the queryset before picking the median ?
                      generally you want it on except for some subtle
                      situations with SQL.

        If the set is empty, returns 0
        """
        qs = self.exclude(**{field + "__isnull": True})
        count = qs.count()

        if not count:
            return datetime.timedelta()

        result = qs.aggregate(
            Sum(field)).get('{}__sum'.format(field)).total_seconds()
        return datetime.timedelta(seconds=int(result / count))


class OwnedModelQuerySet(models.QuerySet):
    def author_query(self, author):
        if self.model.author_path == 'self':
            return Q(pk=author.pk)
        return Q(**{self.model.author_path: author})

    def owner_query(self, owner):
        if self.model.owner_path == 'self':
            return Q(pk=owner.pk)
        return Q(**{self.model.owner_path: owner})

    def orphan_query(self):
        return Q(**{self.model.owner_path: None})

    def from_author(self, author, or_orphan=False):
        """
        :type author: MunchUser
        """
        if or_orphan:
            return self.filter(self.author_query(author) | self.orphan_query())
        return self.filter(self.author_query(author))

    def from_owner(self, owner, or_orphan=False):
        """
        :type owner: Organization
        """
        q = self.owner_query(owner)
        if or_orphan:
            return self.filter(q | self.orphan_query())
        return self.filter(q)


class OptionallyOwnedModelQuerySet(OwnedModelQuerySet):
    def from_author(self, author, or_orphan=True):
        return super().from_author(author, or_orphan)

    def from_owner(self, owner, or_orphan=True):
        return super().from_owner(owner, or_orphan)



from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from django.utils.encoding import force_text
from django.utils.http import urlsafe_base64_decode
from django.contrib.auth import views as auth_views
from django.http.response import HttpResponseRedirect
from django.contrib.auth.password_validation import password_validators_help_text_html  # noqa

from munch.core.utils import get_login_url

from .models import MunchUser


@sensitive_post_parameters()
@never_cache
def password_reset_confirm(request, initial=False, extra_context={}, **kwargs):
    extra_context.update({
        'password_policy': password_validators_help_text_html()})

    response = auth_views.password_reset_confirm(
        request, extra_context=extra_context, **kwargs)

    if isinstance(
            response, HttpResponseRedirect) and response.status_code == 302:
        # Send password change confirmation email
        try:
            uid = force_text(urlsafe_base64_decode(kwargs['uidb64']))
            user = MunchUser.objects.get(pk=uid)
            if initial:
                user.send_invitation_complete_email()
            else:
                user.send_password_reset_complete_email()
        except (TypeError, ValueError, OverflowError, MunchUser.DoesNotExist):
            pass
    return response


def password_reset_complete(request, extra_context={}, **kwargs):
    extra_context.update({'login_url': get_login_url()})
    return auth_views.password_reset_complete(
        request, extra_context=extra_context, **kwargs)



from django.conf.urls import url

from . import views

urlpatterns = [
    url(
        r'^open/(?P<identifier>.+)$',
        views.tracking_open, name='tracking-open'),
    url(
        r'^clicks/m/(?P<identifier>.+)/(?P<link_identifier>.+)$',
        views.tracking_redirect,
        name='tracking-redirect'),
    url(r'^clicks/w/(?P<web_key>.+)/(?P<link_identifier>.+)$',
        views.web_tracking_redirect, name='web-tracking-redirect'),
]



from django.contrib import admin

from .models import Image


@admin.register(Image)
class ImageAdmin(admin.ModelAdmin):
    list_display = (
        'pk', 'organization', 'width', 'upload_date',
        'expiration', 'img_preview')

    def img_preview(self, image):
        return '<a href="{url}" target="_blank">'\
               '<img src="{url}" alt="Preview" height="50" width="50"/>'\
               '</a>'.format(url=image.file.url)
    img_preview.short_description = 'Prévisualisation'
    img_preview.allow_tags = True



from django.conf.urls import url
from django.conf.urls import include

from munch.core.utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register(
    'domains', views.SendingDomainViewSet, base_name='sendingdomain')

urlpatterns = [
    url(
        '^domains/(?P<pk>\d+)/revalidate/$',
        views.SendingDomainRevalidateView.as_view()),
] + router.urls

api_urlpatterns_v1 += [url('', include(urlpatterns, namespace='domains'))]



# -*- coding: utf-8 -*-
from django.db import migrations
from django.core.management.sql import emit_post_migrate_signal

PERMISSIONS = {
    'trackrecord': [
        ('add_trackrecord', 'Can add track record'),
        ('change_trackrecord', 'Can change track record'),
        ('delete_trackrecord', 'Can delete track record'), ],
    'linkmap': [
        ('add_linkmap', 'Can add link map'),
        ('change_linkmap', 'Can change link map'),
        ('delete_linkmap', 'Can delete link map'), ],
}
GROUP_PERMISSIONS = {
    'administrators': {
        'trackrecord': [],
        'linkmap': [],
    },
    'managers': {
        'trackrecord': [],
        'linkmap': [],
    },
    'users': {
        'trackrecord': [],
        'linkmap': [],
    },
    'collaborators': {
        'trackrecord': [],
        'linkmap': [],
    },
}


def update_content_types(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    emit_post_migrate_signal(False, 'default', db_alias)


def load_permissions(apps, schema_editor):
    Group = apps.get_model('auth', 'group')
    Permission = apps.get_model('auth', 'permission')
    ContentType = apps.get_model('contenttypes', 'contenttype')

    # Delete previous permissions
    for model in PERMISSIONS:
        content_type = ContentType.objects.get(
            app_label='tracking', model=model)
        Permission.objects.filter(content_type=content_type).delete()

    # Load permissions
    for model_name, permissions in PERMISSIONS.items():
        for permission_codename, permission_name in permissions:
            content_type = ContentType.objects.get(
                app_label='tracking', model=model_name)
            if not Permission.objects.filter(
                    codename=permission_codename,
                    content_type=content_type).exists():
                Permission.objects.create(
                    name=permission_name,
                    codename=permission_codename,
                    content_type=content_type)

    # Group permissions
    for group_name, models in GROUP_PERMISSIONS.items():
        group, _ = Group.objects.get_or_create(name=group_name)
        for model_name, permissions in models.items():
            content_type = ContentType.objects.get(
                app_label='tracking', model=model_name)
            for codename in permissions:
                permission, _ = Permission.objects.get_or_create(
                    codename=codename, content_type_id=content_type.id)
                group.permissions.add(permission)
                group.save()


class Migration(migrations.Migration):
    # TODO: Add your dependencies here:
    #   dependencies = [('tracking', '0001_initial')]
    dependencies = [('tracking', '0001_initial')]
    operations = [
        migrations.RunPython(update_content_types, reverse_code=None),
        migrations.RunPython(load_permissions, reverse_code=None)]



import factory.django
from faker import Factory as FakerFactory

from ..models import OptOut

faker = FakerFactory.create()


class OptOutFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = OptOut

    origin = OptOut.BY_MAIL
    address = factory.LazyAttribute(lambda x: faker.email())



from django.conf.urls import url
from django.conf.urls import include

from munch.core.utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register(
    'contacts/queues', views.ContactQueueViewSet, base_name='contactqueue')
router.register(
    'contacts/policies',
    views.ContactListPolicyViewSet, base_name='contactlistpolicy')
router.register(
    'contacts/lists', views.ContactListViewSet, base_name='contactlist')
router.register('contacts', views.ContactViewSet, base_name='contact')

urlpatterns = [
    url('^contacts/lists/(?P<contact_list_pk>\d+)/contacts/$',
        views.ContactListContacts.as_view({'post': 'create', 'get': 'list'})),
    url('^contacts/lists/(?P<contact_list_pk>\d+)/merge/$',
        views.ContactListMergeView.as_view()),
] + router.urls

api_urlpatterns_v1 += [url('', include(urlpatterns, namespace='contacts'))]



from django.template import Context
from django.template import Template
from django.utils.encoding import smart_text
from rest_framework.utils import formatting


description_tpl = Template("""
{% if filters %}Results can be filtered/ordered by:
    {% for filter in filters %}`{{ filter }}`
        {% if not forloop.last %}, {% endif %}{% endfor %}.
{% endif %}
{% if default_ordering %}
Default ordering is by `{{ default_ordering }}`.
{% endif %}""")


def get_view_description(view_cls, html=False, instance=None):
    """
    Given a view class, return a textual description to represent the view.
    This name is used in the browsable API, and in OPTIONS responses.
    This function is the default for the `VIEW_DESCRIPTION_FUNCTION` setting.
    """
    # doc can come from the class or from a detail_route or a list_route method
    documented_object = view_cls
    if instance:
        view_method = get_subroute_method(instance)
        if view_method:
            documented_object = view_method

    description = documented_object.__doc__ or ''
    description = formatting.dedent(smart_text(description))

    if hasattr(documented_object, 'filter_class'):
        default_filter = documented_object.filter_class()
        filters = default_filter.filters
        filters_doc = description_tpl.render(Context({
            'filters': filters,
            'default_ordering': default_filter._default_ordering_field()}))
        description += filters_doc

    if html:
        return formatting.markup_description(description)
    return description


def get_subroute_method(viewset):
    """ Detect if we are in a detail_route or list_route and return the relevant
    method, else return None
    """
    if viewset.action:
        view_method = getattr(viewset, viewset.action, None)
        # just an heuristic to check if it's a detail/list route
        if view_method and hasattr(view_method, 'bind_to_methods'):
            return view_method



import django_filters

from munch.core.utils.filters import DRFFilterSet

from ...models import OptOut


class OptOutFilter(DRFFilterSet):
    address = django_filters.CharFilter('address')

    class Meta:
        model = OptOut
        fields = ['address', 'origin', 'creation_date', 'identifier']
        order_by = True



import hashlib
import datetime
from collections import Counter
from collections import defaultdict

import msgpack
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import HStoreField
from django_redis import get_redis_connection

from munch.core.utils import get_mail_by_identifier
from munch.core.mail.models import AbstractMailStatus

conn = get_redis_connection('default')

READ_MUA_PIXEL = 'pixel'
READ_CLICK = 'click'
READ_BROWSER = 'browser'
READ_SOURCES = {
    READ_MUA_PIXEL: 'MUA tracking pixel',
    READ_BROWSER: 'View-in-browser link',
    READ_CLICK: 'Click on a tracked link'}


class TrackRecordQuerySet(models.QuerySet):
    def bulk_create(self, objs, *args, **kwargs):
        for track_record in objs:
            track_record.update_cached_fields()

        created_records = super().bulk_create(objs, *args, **kwargs)
        return created_records

    def count_by_url(self, msg_links, include_any=False, unique=False):
        """
        Returns click count for each URL in the set

        No more than a click per identifier is accounted if unique is True

        :param include_any: adds an extra entry to count mails that have been
                            clicked at least once
        :rtype:             A dict url->int
        """
        click_occurences = self.values(
            'properties', 'identifier').distinct()
        click_lists = defaultdict(list)
        for occ in click_occurences:
            click_lists[occ['identifier']].append(
                msg_links.get(occ['properties']['link']))
        click_lists = dict(click_lists)

        if unique:
            click_lists = {k: list(set(v)) for k, v in click_lists.items()}

        click_counts = Counter()
        for urls in click_lists.values():
            click_counts = click_counts + Counter(urls)
        click_counts = dict(click_counts)

        if include_any:
            click_counts['any'] = len(click_lists.keys())
        return click_counts


class TrackRecord(models.Model):
    identifier = models.CharField(
        max_length=150, db_index=True, verbose_name=_('identifier'))
    kind = models.CharField(
        max_length=50, db_index=True, verbose_name=_('kind'))
    properties = HStoreField(
        null=True, blank=True, verbose_name=_('properties'))
    creation_date = models.DateTimeField(_('creation date'), auto_now_add=True)

    objects = models.Manager.from_queryset(TrackRecordQuerySet)()

    class Meta:
        ordering = ['creation_date']

    def __str__(self):
        return '{} ({})'.format(self.identifier, self.kind)

    @classmethod
    def clear_cache(cls):
        count = 0
        for key in conn.scan_iter('tr:*'):
            count += conn.delete(key)
        return count

    @classmethod
    def get_from_cache(self, identifier, kind=None):
        key = 'tr:{}'.format(identifier)
        records = []
        if kind:
            key += ':{}'.format(kind)
        keys = conn.keys('{}*'.format(key))
        if keys:
            for record in conn.mget(keys):
                record = msgpack.unpackb(record, encoding='utf-8')
                record['creation_date'] = datetime.datetime.fromtimestamp(
                    int(record['creation_date']))
                record.update({'identifier': identifier})
                records.append(record)
            records = sorted(records, key=lambda k: k['creation_date'])
        return records

    def save(self, update_cache=False, *args, **kwargs):
        if not self.creation_date:
            self.creation_date = timezone.now()

        if not self.pk or update_cache:
            self.update_cached_fields()

        if self.kind == 'click' and not TrackRecord.objects.filter(
                identifier=self.identifier, kind='read').exists():
            TrackRecord.objects.create(
                identifier=self.identifier,
                creation_date=self.creation_date,
                kind='read', properties={'source': READ_CLICK})

        super().save(*args, **kwargs)

        self.cache()

    def cache(self):
        return conn.set(
            'tr:{}:{}:{}'.format(self.identifier, self.kind, self.pk),
            msgpack.packb({
                'properties': self.properties,
                'creation_date': self.creation_date.timestamp()}))

    def update_cached_fields(self):
        mail = get_mail_by_identifier(self.identifier)
        delivered_status = mail.statuses.filter(
            status=AbstractMailStatus.DELIVERED).first()
        if delivered_status:
            t = self.creation_date - delivered_status.creation_date
        else:
            t = datetime.timedelta()
        self.properties['reaction_time'] = str(int(t.total_seconds()))


class LinkMapManager(models.Manager):
    def get_identifier(self, link):
        hasher = hashlib.md5()
        if isinstance(link, bytes):
            hasher.update(link)
        else:
            hasher.update(link.encode('utf-8'))
        return hasher.hexdigest()

    def get_or_create(self, defaults=None, **kwargs):
        identifier = self.get_identifier(kwargs.get('link'))
        try:
            return self.get(identifier=identifier), False
        except self.model.DoesNotExist:
            return self.create(
                identifier=identifier, link=kwargs.get('link')), True

    def bulk_create(self, objs, *args, **kwargs):
        # Generate identifiers and deduplicate
        identifiers = []
        to_create = []
        for obj in objs:
            identifier = self.get_identifier(obj.link)
            if identifier not in identifiers:
                identifiers.append(identifier)
                obj.identifier = identifier
                to_create.append(obj)

        existings = self.filter(identifier__in=identifiers)
        existings_identifiers = [obj.identifier for obj in existings]
        objs = []
        for obj in to_create:
            if obj.identifier not in existings_identifiers:
                objs.append(obj)
        return super().bulk_create(objs, *args, **kwargs) + list(existings)


class LinkMap(models.Model):
    identifier = models.CharField(
        max_length=50, db_index=True,
        unique=True, verbose_name=_('identifier'))
    link = models.URLField(verbose_name=_('link'), max_length=500)

    objects = LinkMapManager()

    def __str__(self):
        return '"{}" as "{}"'.format(self.link, self.identifier)

    def save(self, *args, **kwargs):
        if not self.identifier:
            self.identifier = LinkMap.objects.get_identifier(self.link)
        super().save(*args, **kwargs)



from django.test import TestCase

from munch.apps.users.tests.factories import UserFactory
from munch.apps.transactional.models import Mail as TransactionalMail

from ..models import RawMail
from ..tasks import purge_raw_mail


class PurgeRawMailTestCase(TestCase):
    def test_all(self):
        user = UserFactory()

        raw_mail_01, _ = RawMail.objects.get_or_create(content='test 01')
        raw_mail_02, _ = RawMail.objects.get_or_create(content='test_02')

        mail_01 = TransactionalMail.objects.create(
            identifier='1', recipient='foo@bar', headers={'From': 'bar@foo'},
            sender='bar@foo', message=raw_mail_01, author=user)
        mail_02 = TransactionalMail.objects.create(
            identifier='2', recipient='bar@foo', headers={'From': 'foo@bar'},
            sender='foo@bar', message=raw_mail_02, author=user)

        purge_raw_mail()

        self.assertEqual(RawMail.objects.count(), 2)

        mail_01.message = None
        mail_01.save()

        purge_raw_mail()
        self.assertEqual(RawMail.objects.count(), 1)

        mail_03 = TransactionalMail.objects.create(
            identifier='3', recipient='bar@foo', headers={'From': 'foo@bar'},
            sender='foo@bar', message=raw_mail_02, author=user)

        mail_02.message = None
        mail_02.save()

        purge_raw_mail()
        self.assertEqual(RawMail.objects.count(), 1)



import factory
import factory.django
from django.utils import timezone
from django.contrib.auth.models import Group
from faker import Factory as FakerFactory

from munch.core.mail.utils import mk_base64_uuid

from ..models import MunchUser
from ..models import Organization
from ..models import APIApplication
from ..models import SmtpApplication


faker = FakerFactory.create()


class OrganizationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Organization

    name = factory.LazyAttribute(lambda x: faker.sentence(nb_words=3))
    contact_email = factory.LazyAttribute(lambda x: faker.email())


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MunchUser

    organization = factory.SubFactory(OrganizationFactory)
    identifier = factory.LazyAttribute(lambda x: faker.email())
    last_login = factory.LazyAttribute(lambda x: timezone.now())
    secret = factory.LazyAttribute(lambda x: mk_base64_uuid())
    is_active = True
    password = factory.PostGenerationMethodCall('set_password', 'password')

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for group in extracted:
                if group in [
                        'users', 'collaborators',
                        'managers', 'administrators']:
                    self.groups.add(Group.objects.filter(name=group)[0])


class SmtpApplicationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = SmtpApplication

    identifier = factory.LazyAttribute(lambda x: faker.sentence(nb_words=2))
    author = factory.SubFactory(UserFactory)


class APIApplicationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = APIApplication

    identifier = factory.LazyAttribute(lambda x: faker.sentence(nb_words=2))
    author = factory.SubFactory(UserFactory)



from django.conf import settings
from slimta.policy import QueuePolicy

from munch.core.mail.utils import mk_msgid


class Remove(QueuePolicy):
    """
    Remove internal headers to avoid spoofing
    """
    def apply(self, envelope):
        for header in settings.TRANSACTIONAL.get('HEADERS_TO_REMOVE'):
            del envelope.headers[header]


class AddMessageIdHeader(QueuePolicy):
    """
    Add a Message-ID header if absent with a custom msgid domain
    """
    def apply(self, envelope):
        if 'message-id' not in envelope.headers:
            envelope.headers['Message-Id'] = mk_msgid()



from munch.core.mail.exceptions import SoftFailure


def get_envelope(identifier):
    """ This method is only for email Backend """
    from .models import Mail
    try:
        return Mail.objects.get(identifier=identifier).as_envelope()
    except Mail.DoesNotExist as exc:
        raise SoftFailure(exc)


def get_envelope_from_identifier(identifier, only_headers=False):
    """ This method is for campaigns app """
    from .models import Mail
    return Mail.objects.get(identifier=identifier).as_envelope(
        must_raise=not only_headers)



from datetime import datetime

import pytz
import factory.django
from django.conf import settings
from faker import Factory as FakerFactory

from munch.core.mail.utils import mk_base64_uuid
from munch.apps.domains.models import SendingDomain

from ..models import Mail
from ..models import Message
from ..models import MailStatus
from ..models import PreviewMail
from ..models import get_base_mail_identifier

faker = FakerFactory.create()


class MessageFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Message

    class Params:
        html_content = '{} {}'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'],
            settings.HOSTED['WEB_LINK_PLACEHOLDER'])

    name = factory.LazyAttribute(lambda x: faker.sentence(nb_words=3))
    creation_date = factory.LazyAttribute(lambda x: datetime.now(pytz.UTC))
    sender_email = factory.LazyAttribute(lambda x: faker.email())
    sender_name = factory.LazyAttribute(lambda x: faker.name())
    identifier = factory.LazyAttribute(lambda x: mk_base64_uuid())
    html = factory.LazyAttribute(lambda o: '<body>{} {}</body>'.format(
        o.html_content, faker.text(max_nb_chars=100)))

    @factory.post_generation
    def auto_create_sending_domain(
            self, create, extracted, create_sending_domain=True):
        if not create:
            return

        if create_sending_domain:
            SendingDomain.objects.create(
                name=self.name, organization=self.author.organization)


class MailFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Mail

    message = factory.SubFactory(MessageFactory)
    curstatus = 'unknown'
    creation_date = factory.LazyAttribute(lambda x: datetime.now(pytz.UTC))
    identifier = factory.LazyAttribute(lambda x: get_base_mail_identifier())
    recipient = factory.LazyAttribute(lambda x: faker.email())


class MailStatusFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MailStatus

    creation_date = factory.LazyAttribute(lambda x: datetime.now(pytz.UTC))


class PreviewMailFactory(MailFactory):
    class Meta:
        model = PreviewMail



from munch.core.utils.permissions import MunchResourcePermission


class MunchUserPermission(MunchResourcePermission):
    app_name = 'users'
    model_name = 'munchuser'

    def has_object_permission(self, request, view, obj):
        """ Check that the user isn't deleting itself
        """
        if (request.user == obj) and (request.method == 'DELETE'):
            return False
        return super().has_object_permission(request, view, obj)


class OrganizationPermission(MunchResourcePermission):
    app_name = 'users'
    model_name = 'organization'

    def has_object_permission(self, request, view, obj):
        if obj in request.user.organization.children.all():
            return True
        return super().has_object_permission(request, view, obj)

    def has_permission(self, request, view):
        if request.method in ['POST', 'PUT', 'PATCH', 'DELETE']:
            user = request.user
            permissions = [
                'users.add_child_organization',
                'users.delete_child_organization']
            if user.is_authenticated():
                if user.groups.filter(name='administrators').exists():
                    if user.has_perms(permissions):
                        return True
        return super().has_permission(request, view)



from django.utils.dateparse import parse_duration
from rest_framework import generics
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from munch.core.utils.permissions import IsAuthenticatedOrPreflight

from ...models import Image
from ...models import UploadDuplicateError
from .serializers import ImageSerializer


@api_view(('GET',))
def api_root(request, format=None):
    return Response({})


class ImageDetail(generics.RetrieveAPIView):
    queryset = Image.objects.all()
    serializer_class = ImageSerializer
    permission_classes = tuple()
    versioning_class = None


class ImageCreate(generics.CreateAPIView):
    queryset = Image.objects.all()
    serializer_class = ImageSerializer
    permission_classes = (IsAuthenticatedOrPreflight,)
    http_method_names = ('post', 'options')
    versioning_class = None

    def create(self, request, *args, **kwargs):
        try:
            return super().create(request, *args, **kwargs)
        except UploadDuplicateError as e:
            dup = Image.objects.get(pk=e.instance.hash)
            # this usage of parse_duration is safe, because we already
            # validated the data (exception is thrown by the Model class, so
            # the validator passed).
            if 'expiration' in request.data:
                dup.expiration = parse_duration(request.data['expiration'])
            dup.save()

            serializer = self.get_serializer(dup)
            headers = self.get_success_headers(serializer.data)
            return Response(serializer.data, status=status.HTTP_201_CREATED,
                            headers=headers)



from django.contrib import admin
from .models import AbuseNotification


class AbuseNotificationAdmin(admin.ModelAdmin):
    list_display = ['mail', 'contact_email', 'date']
    raw_id_fields = ['mail']
    search_fields = ['mail__recipient']


admin.site.register(AbuseNotification, AbuseNotificationAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-05-27 08:15
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('campaigns', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AbuseNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now=True, verbose_name='date')),
                ('first_name', models.CharField(max_length=60, verbose_name='firstname')),
                ('second_name', models.CharField(max_length=60, verbose_name='lastname')),
                ('organization', models.CharField(blank=True, max_length=100, verbose_name='organization')),
                ('contact_phone', models.CharField(blank=True, max_length=50, validators=[django.core.validators.RegexValidator(message='Phone number format must be like "+33123456789"', regex='^\\+?1?[\\d\\.\\s]{9,20}$')], verbose_name='phone number')),
                ('contact_mail', models.EmailField(blank=True, max_length=254, verbose_name='email')),
                ('comment', models.TextField(blank=True, verbose_name='additional information')),
                ('mail', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='campaigns.Mail', verbose_name='contact email report')),
            ],
            options={
                'verbose_name_plural': 'abuse reports',
                'verbose_name': 'abuse report',
            },
        ),
    ]



from django.contrib.auth import get_user_model
from django.utils.encoding import force_text
from django.utils.http import urlsafe_base64_decode
from rest_framework import status
from rest_framework import viewsets
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import detail_route
from rest_framework_jwt.views import JSONWebTokenAPIView

from munch.apps.campaigns.models import Mail
from munch.apps.campaigns.models import OptOut
from munch.apps.optouts.api.v1.filters import OptOutFilter
from munch.apps.optouts.api.v1.serializers import OptOutSerializer
from munch.core.utils.views import filtered
from munch.core.utils.views import paginated
from munch.core.utils.views import NestedView
from munch.core.utils.views import MunchModelViewSetMixin
from munch.core.utils.views import OrganizationOwnedViewSetMixin

from ...models import MunchUser
from ...models import Organization
from ...models import APIApplication
from ...models import SmtpApplication
from ...tokens import MunchUserTokenGenerator
from .filters import MunchUserFilter
from .permissions import MunchUserPermission
from .permissions import OrganizationPermission
from .serializers import MeSerializer
from .serializers import MunchUserSerializer
from .serializers import OrganizationSerializer
from .serializers import APIApplicationSerializer
from .serializers import SmtpApplicationSerializer
from .serializers import PasswordResetInitSerializer
from .serializers import PasswordChangeSerializer
from .serializers import MunchJSONWebTokenSerializer
from .serializers import InvitationSerializer


class MeAPIView(APIView):
    def get(self, request, format=None):
        if request.user.is_authenticated():
            serializer = MeSerializer(
                request.user, context={'request': request})
            return Response(serializer.data)
        return Response({}, status=status.HTTP_403_FORBIDDEN)


class OrganizationViewSet(
        OrganizationOwnedViewSetMixin, viewsets.ModelViewSet):
    model = Organization
    serializer_class = OrganizationSerializer
    permission_classes = [OrganizationPermission]

    @detail_route(methods=['get'])
    @paginated(OptOutSerializer)
    @filtered(OptOutFilter)
    def opt_outs(self, request, pk, format=None):
        identifiers = Mail.objects.filter(
            message__author__organization=self.get_object()).values_list(
            'identifier', flat=True)
        return OptOut.objects.filter(identifier__in=identifiers)

    @detail_route(methods=['get'])
    def children(self, request, pk, format=None):
        childs = Organization.objects.filter(parent=self.get_object())
        serializer = OrganizationSerializer(
            childs, many=True, context={'request': request})
        return Response(serializer.data)

    @detail_route(methods=['post'])
    def invite_user(self, request, pk, format=None):
        if not request.user.has_perm('users.invite_user_organization'):
            return Response(status=401)
        serializer = InvitationSerializer(
            data=request.data,
            context={'request': request, 'organization_id': pk})
        if serializer.is_valid():
            user = MunchUser(
                identifier=serializer.validated_data.get('identifier'),
                organization_id=pk, invited_by=request.user)
            user.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(
                serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class ChangePasswordView(NestedView):
    """
    Change human user password

    Can be done by the user itself or an admin of same organization.
    """

    parent_model = MunchUser
    serializer_class = PasswordChangeSerializer

    def post(self, request, user_pk):
        # Might be in HumanUserViewSet but here for comodity
        # (see HALLinksField)
        user = self.get_parent_object(user_pk)
        serializer = self.get_serializer(user=user, data=request.data)

        if serializer.is_valid():
            user.set_password(serializer.validated_data['new_password'])
            user.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(
                serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class RegenSecretView(NestedView):
    """
    Regenerate an user key

    returns the generated key
    """
    parent_model = MunchUser

    def post(self, request, user_pk):
        user = self.get_parent_object(user_pk)
        user.regen_secret()
        user.save()
        return Response(user.secret)


class MunchUserViewSet(
        MunchModelViewSetMixin,
        OrganizationOwnedViewSetMixin,
        viewsets.ReadOnlyModelViewSet):
    model = MunchUser
    serializer_class = MunchUserSerializer
    filter_class = MunchUserFilter
    permission_classes = [MunchUserPermission]


class PasswordResetInitView(APIView):
    """ That view allows to initiate a password reset

    It will always issue HTTP 201 so an attacker can't find if a
    given email is a munch user or not.
    """
    permission_classes = []

    def post(self, request, format=None):
        serializer = PasswordResetInitSerializer(data=request.data)
        if serializer.is_valid():
            # email = request.data['email']
            email = serializer.validated_data['email']
            try:
                user = MunchUser.objects.get(identifier=email)
            except MunchUser.DoesNotExist:
                # say nothing
                pass
            else:
                user.send_password_reset_email()
            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(
                serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class PasswordResetSetView(APIView):
    """ That view allows to set a new password given an uid and a token

    (the token is received by email)
    """
    # That's open to everybody
    permission_classes = []

    def post(self, request, format=None):
        UserModel = get_user_model()
        uidb64 = request.data['uid']
        token = request.data['token']
        new_password = request.data['new_password']

        token_generator = MunchUserTokenGenerator()
        try:
            # urlsafe_base64_decode() decodes to bytestring on Python 3
            uid = force_text(urlsafe_base64_decode(uidb64))
            user = MunchUser.objects.get(pk=uid)
        except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
            user = None

        if token_generator.check_token(user, token):
            user.set_password(new_password)
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(status=status.HTTP_401_UNAUTHORIZED)


class ObtainJSONWebToken(JSONWebTokenAPIView):
    """
    API View that receives a POST with a user's username and password.

    Returns a JSON Web Token that can be used for authenticated requests.
    """
    serializer_class = MunchJSONWebTokenSerializer


class APIApplicationViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = APIApplication
    serializer_class = APIApplicationSerializer

    @detail_route(methods=['post'])
    def regen_secret(self, request, pk=None):
        application = self.get_object()
        application.regen_secret()
        application.save()
        return Response(application.secret)


class SmtpApplicationViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = SmtpApplication
    serializer_class = SmtpApplicationSerializer

    @detail_route(methods=['post'])
    def regen_credentials(self, request, pk=None):
        application = self.get_object()
        application.regen_credentials()
        application.save()
        return Response(
            {"username": application.username, "secret": application.secret})



from munch.core.utils.filters import DRFFilterSet

from ...models import SendingDomain


class SendingDomainFilter(DRFFilterSet):
    class Meta:
        model = SendingDomain
        fields = ['name', 'dkim_status']
        order_by = True
